from .base import DiagnosticBase
from .diagnostic_8gig import diagnostic8Gig
from .diagnostic_20gig import diagnostic20Gig
from .diagnostic_network import diagnosticNetwork


class allDiagnostics(DiagnosticBase):
    def getName(self):
        """
        Returns the human-readable name of the diagnostic
        """
        return "Run all available diagnostics"

    def getDescription(self):
        """
        Returns a description of what the diagnostic does
        """
        return "This diagnostic runs all available diagnostics in sequence."

    def run(self, logger, args=[]):
        """
        Runs the diagnostic
        """

        # Run all available diagnostics in turn, storing the results
        results = []
        diagnostics = [
            diagnostic8Gig(),
            diagnostic20Gig(),
            diagnosticNetwork(),
        ]
        for index, diagnostic in enumerate(diagnostics):
            # Run the diagnostic and report its result
            logger.info(
                '[all] Running individual diagnostic: "{}"'.format(
                    diagnostic.getName()
                ),
                True,
            )
            results.append(diagnostic.run(logger))
            logger.info(
                "[all] Individual diagnostic result: {}".format(
                    "passed" if results[-1] == True else "failed"
                ),
                False,
            )

            # Print a newline after the last diagnostic has run
            if index == len(diagnostics) - 1:
                print()

        # Only report success if all diagnostics succeeded
        return False not in results



#!/usr/bin/env python3
import glob, os, shutil, sys
from os.path import basename, dirname, exists, join


# Logs a message to stderr
def log(message):
    print(message, file=sys.stderr)
    sys.stderr.flush()


# Extracts the files and directories for the specified component and moves them to a separate output directory
def extractComponent(inputDir, outputDir, component, description, items):
    # Print progress output
    log("\nExtracting {}...".format(description))

    # Create the output directory for the component if it doesn't already exist
    componentDir = join(outputDir, component)
    os.makedirs(outputDir, exist_ok=True)

    # Move each file and directory for the component to the output directory
    for item in items:
        # Verify that the item exists
        if not exists(item):
            log("Skipping non-existent item: {}".format(item))
            continue

        # Print progress output
        log("Moving: {}".format(item))

        # Ensure the parent directory for the item exists in the output directory
        parent = dirname(item).replace(inputDir, componentDir)
        os.makedirs(parent, exist_ok=True)

        # Perform the move
        shutil.move(item, join(parent, basename(item)))


# Retrieve the path to the root directory of the Installed Build
rootDir = sys.argv[1]

# Retrieve the path to the root output directory for extracted components and ensure it exists
outputDir = sys.argv[2]
os.makedirs(outputDir, exist_ok=True)

# Extract the DDC
ddc = [join(rootDir, "Engine", "DerivedDataCache", "Compressed.ddp")]
extractComponent(rootDir, outputDir, "DDC", "Derived Data Cache (DDC)", ddc)

# Extract debug symbols
symbolFiles = glob.glob(join(rootDir, "**", "*.debug"), recursive=True) + glob.glob(
    join(rootDir, "**", "*.sym"), recursive=True
)
extractComponent(rootDir, outputDir, "DebugSymbols", "debug symbols", symbolFiles)

# Extract template projects and samples
subdirs = [join(rootDir, subdir) for subdir in ["FeaturePacks", "Samples", "Templates"]]
extractComponent(
    rootDir, outputDir, "TemplatesAndSamples", "template projects and samples", subdirs
)



import docker, os, platform, requests, shutil, subprocess, sys
from .infrastructure import *


# Runs a command without displaying its output and returns the exit code
def _runSilent(command):
    result = SubprocessUtils.capture(command, check=False)
    return result.returncode


# Performs setup for Linux hosts
def _setupLinux():
    # Pull the latest version of the Alpine container image
    alpineImage = "alpine:latest"
    SubprocessUtils.capture(["docker", "pull", alpineImage])

    # Start the credential endpoint with blank credentials
    endpoint = CredentialEndpoint("", "")
    endpoint.start()

    try:
        # Run an Alpine container to see if we can access the host port for the credential endpoint
        SubprocessUtils.capture(
            [
                "docker",
                "run",
                "--rm",
                alpineImage,
                "wget",
                "--timeout=1",
                "--post-data=dummy",
                "http://{}:9876".format(NetworkUtils.hostIP()),
            ],
            check=True,
        )

        # If we reach this point then the host port is accessible
        print("No firewall configuration required.")

    except:
        # The host port is blocked, so we need to perform firewall configuration
        print("Creating firewall rule for credential endpoint...")

        # Create the firewall rule
        subprocess.run(
            ["iptables", "-I", "INPUT", "-p", "tcp", "--dport", "9876", "-j", "ACCEPT"],
            check=True,
        )

        # Ensure the firewall rule persists after reboot
        # (Requires the `iptables-persistent` service to be installed and running)
        os.makedirs("/etc/iptables", exist_ok=True)
        subprocess.run("iptables-save > /etc/iptables/rules.v4", shell=True, check=True)

        # Inform users of the `iptables-persistent` requirement
        print(
            "Firewall rule created. Note that the `iptables-persistent` service will need to"
        )
        print("be installed for the rule to persist after the host system reboots.")

    finally:
        # Stop the credential endpoint
        endpoint.stop()


# Performs setup for Windows Server hosts
def _setupWindowsServer():
    # Check if we need to configure the maximum image size
    requiredLimit = WindowsUtils.requiredSizeLimit()
    if DockerUtils.maxsize() < requiredLimit:
        # Attempt to stop the Docker daemon
        print("Stopping the Docker daemon...")
        subprocess.run(["sc.exe", "stop", "docker"], check=True)

        # Attempt to set the maximum image size
        print("Setting maximum image size to {}GB...".format(requiredLimit))
        config = DockerUtils.getConfig()
        sizeOpt = "size={}GB".format(requiredLimit)
        if "storage-opts" in config:
            config["storage-opts"] = list(
                [
                    o
                    for o in config["storage-opts"]
                    if o.lower().startswith("size=") == False
                ]
            )
            config["storage-opts"].append(sizeOpt)
        else:
            config["storage-opts"] = [sizeOpt]
        DockerUtils.setConfig(config)

        # Attempt to start the Docker daemon
        print("Starting the Docker daemon...")
        subprocess.run(["sc.exe", "start", "docker"], check=True)

    else:
        print("Maximum image size is already correctly configured.")

    # Determine if we need to configure Windows firewall
    ruleName = "Open TCP port 9876 for ue4-docker credential endpoint"
    ruleExists = (
        _runSilent(
            [
                "netsh",
                "advfirewall",
                "firewall",
                "show",
                "rule",
                "name={}".format(ruleName),
            ]
        )
        == 0
    )
    if ruleExists == False:
        # Add a rule to ensure Windows firewall allows access to the credential helper from our containers
        print("Creating firewall rule for credential endpoint...")
        subprocess.run(
            [
                "netsh",
                "advfirewall",
                "firewall",
                "add",
                "rule",
                "name={}".format(ruleName),
                "dir=in",
                "action=allow",
                "protocol=TCP",
                "localport=9876",
            ],
            check=True,
        )

    else:
        print("Firewall rule for credential endpoint is already configured.")


def setup():
    # We don't currently support auto-config for VM-based containers
    if platform.system() == "Darwin" or (
        platform.system() == "Windows" and WindowsUtils.isWindowsServer() == False
    ):
        print(
            "Manual configuration is required under Windows 10 and macOS. Automatic configuration is not available."
        )
        return

    # Perform setup based on the host system type
    if platform.system() == "Linux":
        _setupLinux()
    else:
        _setupWindowsServer()



import docker, fnmatch, humanfriendly, itertools, json, logging, os, platform, re
from docker.models.containers import Container

from .FilesystemUtils import FilesystemUtils


class DockerUtils(object):
    @staticmethod
    def installed():
        """
        Determines if Docker is installed
        """
        try:
            return (DockerUtils.version() is not None), None
        except Exception as e:
            logging.debug(str(e))
            return False, e

    @staticmethod
    def version():
        """
        Retrieves the version information for the Docker daemon
        """
        client = docker.from_env()
        return client.version()

    @staticmethod
    def info():
        """
        Retrieves the system information as produced by `docker info`
        """
        client = docker.from_env()
        return client.info()

    @staticmethod
    def exists(name):
        """
        Determines if the specified image exists
        """
        client = docker.from_env()
        try:
            image = client.images.get(name)
            return True
        except:
            return False

    @staticmethod
    def build(tags: [str], context: str, args: [str]) -> [str]:
        """
        Returns the `docker build` command to build an image
        """
        tagArgs = [["-t", tag] for tag in tags]
        return (
            ["docker", "build"]
            + list(itertools.chain.from_iterable(tagArgs))
            + [context]
            + args
        )

    @staticmethod
    def buildx(tags: [str], context: str, args: [str], secrets: [str]) -> [str]:
        """
        Returns the `docker buildx` command to build an image with the BuildKit backend
        """
        tagArgs = [["-t", tag] for tag in tags]
        return (
            ["docker", "build"]
            + list(itertools.chain.from_iterable(tagArgs))
            + [context]
            + ["--progress=plain"]
            + args
            + list(itertools.chain.from_iterable([["--secret", s] for s in secrets]))
        )

    @staticmethod
    def pull(image):
        """
        Returns the `docker pull` command to pull an image from a remote registry
        """
        return ["docker", "pull", image]

    @staticmethod
    def start(image, command, **kwargs):
        """
        Starts a container in a detached state and returns the container handle
        """
        client = docker.from_env()
        return client.containers.run(image, command, detach=True, **kwargs)

    @staticmethod
    def create(image: str, **kwargs) -> Container:
        """
        Creates a stopped container for specified image name and returns the container handle
        """
        client = docker.from_env()
        return client.containers.create(image, **kwargs)

    @staticmethod
    def configFilePath():
        """
        Returns the path to the Docker daemon configuration file under Windows
        """
        return "{}\\Docker\\config\\daemon.json".format(os.environ["ProgramData"])

    @staticmethod
    def getConfig():
        """
        Retrieves and parses the Docker daemon configuration file under Windows
        """
        configPath = DockerUtils.configFilePath()
        if os.path.exists(configPath) == True:
            with open(configPath) as configFile:
                return json.load(configFile)

        return {}

    @staticmethod
    def setConfig(config):
        """
        Writes new values to the Docker daemon configuration file under Windows
        """
        configPath = DockerUtils.configFilePath()
        with open(configPath, "w") as configFile:
            configFile.write(json.dumps(config))

    @staticmethod
    def maxsize():
        """
        Determines the configured size limit (in GB) for Windows containers
        """
        if platform.system() != "Windows":
            return -1

        config = DockerUtils.getConfig()
        if "storage-opts" in config:
            sizes = [
                opt.replace("size=", "")
                for opt in config["storage-opts"]
                if "size=" in opt
            ]
            if len(sizes) > 0:
                return humanfriendly.parse_size(sizes[0]) / 1000000000

        # The default limit on image size is 20GB
        # (https://docs.microsoft.com/en-us/visualstudio/install/build-tools-container-issues)
        return 20.0

    @staticmethod
    def listImages(tagFilter=None, filters={}, all=False):
        """
        Retrieves the details for each image matching the specified filters
        """

        # Retrieve the list of images matching the specified filters
        client = docker.from_env()
        images = client.images.list(filters=filters, all=all)

        # Apply our tag filter if one was specified
        if tagFilter is not None:
            images = [
                i
                for i in images
                if len(i.tags) > 0 and len(fnmatch.filter(i.tags, tagFilter)) > 0
            ]

        return images

    @staticmethod
    def exec(container, command: [str], **kwargs):
        """
        Executes a command in a container returned by `DockerUtils.start()` and returns the output
        """
        result, output = container.exec_run(command, **kwargs)
        if result is not None and result != 0:
            container.stop()
            raise RuntimeError(
                "Failed to run command {} in container. Process returned exit code {} with output: {}".format(
                    command, result, output
                )
            )

        return output

    @staticmethod
    def execMultiple(container, commands: [[str]], **kwargs):
        """
        Executes multiple commands in a container returned by `DockerUtils.start()`
        """
        for command in commands:
            DockerUtils.exec(container, command, **kwargs)

    @staticmethod
    def injectPostRunMessage(dockerfile, platform, messageLines):
        """
        Injects the supplied message at the end of each RUN directive in the specified Dockerfile
        """

        # Generate the `echo` command for each line of the message
        prefix = "echo." if platform == "windows" else "echo '"
        suffix = "" if platform == "windows" else "'"
        echoCommands = "".join(
            [" && {}{}{}".format(prefix, line, suffix) for line in messageLines]
        )

        # Read the Dockerfile contents and convert all line endings to \n
        contents = FilesystemUtils.readFile(dockerfile)
        contents = contents.replace("\r\n", "\n")

        # Determine the escape character for the Dockerfile
        escapeMatch = re.search("#[\\s]*escape[\\s]*=[\\s]*([^\n])\n", contents)
        escape = escapeMatch[1] if escapeMatch is not None else "\\"

        # Identify each RUN directive in the Dockerfile
        runMatches = re.finditer(
            "^RUN(.+?[^{}])\n".format(re.escape(escape)),
            contents,
            re.DOTALL | re.MULTILINE,
        )
        if runMatches is not None:
            for match in runMatches:
                # Append the `echo` commands to the directive
                contents = contents.replace(
                    match[0], "RUN{}{}\n".format(match[1], echoCommands)
                )

        # Write the modified contents back to the Dockerfile
        FilesystemUtils.writeFile(dockerfile, contents)



import subprocess


class VerboseCalledProcessError(RuntimeError):
    """'
    A verbose wrapper for `subprocess.CalledProcessError` that prints stdout and stderr
    """

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __str__(self):
        return "{}\nstdout: {}\nstderr: {}".format(
            self.wrapped, self.wrapped.output, self.wrapped.stderr
        )


class SubprocessUtils(object):
    @staticmethod
    def extractLines(output):
        """
        Extracts the individual lines from the output of a child process
        """
        return output.decode("utf-8").replace("\r\n", "\n").strip().split("\n")

    @staticmethod
    def capture(command, check=True, **kwargs):
        """
        Executes a child process and captures its output.

        If the child process fails and `check` is True then a verbose exception will be raised.
        """
        try:
            return subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                check=check,
                **kwargs,
            )
        except subprocess.CalledProcessError as e:
            raise VerboseCalledProcessError(e) from None

    @staticmethod
    def run(command, check=True, **kwargs):
        """
        Executes a child process.

        If the child process fails and `check` is True then a verbose exception will be raised.
        """
        return SubprocessUtils.capture(command, check, **kwargs)



from ..infrastructure import DockerUtils, WindowsUtils
from .base import DiagnosticBase

import argparse, os, platform
from os.path import abspath, dirname, join


class diagnostic8Gig(DiagnosticBase):
    # The tag we use for built images
    IMAGE_TAG = "adamrehn/ue4-docker/diagnostics:8gig"

    def __init__(self):
        # Setup our argument parser so we can use its help message output in our description text
        self._parser = argparse.ArgumentParser(prog="ue4-docker diagnostics 8gig")
        self._parser.add_argument(
            "--linux",
            action="store_true",
            help="Use Linux containers under Windows hosts (useful when testing Docker Desktop or LCOW support)",
        )
        self._parser.add_argument(
            "--random",
            action="store_true",
            help="Create a file filled with random bytes instead of zeroes under Windows",
        )
        self._parser.add_argument(
            "--isolation",
            default=None,
            choices=["hyperv", "process"],
            help="Override the default isolation mode when testing Windows containers",
        )
        self._parser.add_argument(
            "-basetag",
            default=None,
            choices=WindowsUtils.getKnownBaseTags(),
            help="Override the default base image tag when testing Windows containers",
        )

    def getName(self):
        """
        Returns the human-readable name of the diagnostic
        """
        return "Check for Docker 8GiB filesystem layer bug"

    def getDescription(self):
        """
        Returns a description of what the diagnostic does
        """
        return "\n".join(
            [
                "This diagnostic determines if the Docker daemon suffers from one of the 8GiB filesystem",
                "layer bugs reported at https://github.com/moby/moby/issues/37581 (affects all platforms)",
                "or https://github.com/moby/moby/issues/40444 (affects Windows containers only)",
                "",
                "#37581 was fixed in Docker CE 18.09.0 and #40444 was fixed in Docker CE 20.10.0",
                "",
                self._parser.format_help(),
            ]
        )

    def getPrefix(self):
        """
        Returns the short name of the diagnostic for use in log output
        """
        return "8gig"

    def run(self, logger, args=[]):
        """
        Runs the diagnostic
        """

        # Parse our supplied arguments
        args = self._parser.parse_args(args)

        # Determine which image platform we will build the Dockerfile for (default is the host platform unless overridden)
        containerPlatform = (
            "linux"
            if args.linux == True or platform.system().lower() != "windows"
            else "windows"
        )

        # Verify that the user isn't trying to test Windows containers under Windows 10 when in Linux container mode (or vice versa)
        try:
            self._checkPlatformMistmatch(logger, containerPlatform)
        except RuntimeError:
            return False

        # Set our build arguments when testing Windows containers
        buildArgs = (
            self._generateWindowsBuildArgs(logger, args.basetag, args.isolation)
            if containerPlatform == "windows"
            else []
        )

        # Attempt to build the Dockerfile
        logger.action(
            "[8gig] Attempting to build an image with an 8GiB filesystem layer...",
            False,
        )
        built = self._buildDockerfile(
            logger, containerPlatform, diagnostic8Gig.IMAGE_TAG, buildArgs
        )

        # Inform the user of the outcome of the diagnostic
        if built == True:
            logger.action(
                "[8gig] Diagnostic succeeded! The Docker daemon can build images with 8GiB filesystem layers.\n"
            )
        else:
            logger.error(
                "[8gig] Diagnostic failed! The Docker daemon cannot build images with 8GiB filesystem layers.\n",
                True,
            )

        return built



import argparse, subprocess, sys
from .infrastructure import *


def _isIntermediateImage(image):
    sentinel = "com.adamrehn.ue4-docker.sentinel"
    labels = image.attrs["ContainerConfig"]["Labels"]
    return labels is not None and sentinel in labels


def _cleanMatching(cleaner, filter, tag, dryRun):
    tagSuffix = ":{}".format(tag) if tag is not None else "*"
    matching = DockerUtils.listImages(tagFilter=filter + tagSuffix)
    cleaner.cleanMultiple([image.tags[0] for image in matching], dryRun)


def clean():
    # Create our logger to generate coloured output on stderr
    logger = Logger(prefix="[{} clean] ".format(sys.argv[0]))

    # Our supported command-line arguments
    parser = argparse.ArgumentParser(
        prog="{} clean".format(sys.argv[0]),
        description="Cleans built container images. "
        + "By default, only dangling intermediate images leftover from ue4-docker multi-stage builds are removed.",
    )
    parser.add_argument(
        "-tag", default=None, help="Only clean images with the specified tag"
    )
    parser.add_argument("--source", action="store_true", help="Clean ue4-source images")
    parser.add_argument(
        "--all", action="store_true", help="Clean all ue4-docker images"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Print docker commands instead of running them",
    )
    parser.add_argument(
        "--prune", action="store_true", help="Run `docker system prune` after cleaning"
    )

    # Parse the supplied command-line arguments
    args = parser.parse_args()

    # Create our image cleaner
    cleaner = ImageCleaner(logger)

    # Remove any intermediate images leftover from our multi-stage builds
    dangling = DockerUtils.listImages(filters={"dangling": True})
    dangling = [image.id for image in dangling if _isIntermediateImage(image)]
    cleaner.cleanMultiple(dangling, args.dry_run)

    # If requested, remove ue4-source images
    if args.source == True:
        _cleanMatching(
            cleaner,
            GlobalConfiguration.resolveTag("ue4-source"),
            args.tag,
            args.dry_run,
        )

    # If requested, remove everything
    if args.all == True:
        _cleanMatching(
            cleaner, GlobalConfiguration.resolveTag("ue4-*"), args.tag, args.dry_run
        )

    # If requested, run `docker system prune`
    if args.prune == True:
        logger.action("Running `docker system prune`...")
        pruneCommand = ["docker", "system", "prune", "-f"]
        if args.dry_run == True:
            print(pruneCommand)
        else:
            subprocess.call(pruneCommand)



import contextlib
import io
import logging
import os
import shutil
import sys
import tempfile

import docker
from docker.models.containers import Container


class ContainerUtils(object):
    """
    Provides functionality related to Docker containers
    """

    @staticmethod
    @contextlib.contextmanager
    def automatically_stop(container: Container, timeout: int = 1):
        """
        Context manager to automatically stop a container returned by `ContainerUtils.start_for_exec()`
        """
        try:
            yield container
        finally:
            logging.info("Stopping Docker container {}...".format(container.short_id))
            container.stop(timeout=timeout)

    @staticmethod
    def copy_from_host(
        container: Container, host_path: str, container_path: str
    ) -> None:
        """
        Copies a file or directory from the host system to a container returned by `ContainerUtils.start_for_exec()`.

        `host_path` is the absolute path to the file or directory on the host system.

        `container_path` is the absolute path to the directory in the container where the copied file(s) will be placed.
        """

        # If the host path denotes a file rather than a directory, copy it to a temporary directory
        # (If the host path is a directory then we create a no-op context manager to use in our `with` statement below)
        tempDir = contextlib.suppress()
        if os.path.isfile(host_path):
            tempDir = tempfile.TemporaryDirectory()
            shutil.copy2(
                host_path, os.path.join(tempDir.name, os.path.basename(host_path))
            )
            host_path = tempDir.name

        # Automatically delete the temporary directory if we created one
        with tempDir:
            # Create a temporary file to hold the .tar archive data
            with tempfile.NamedTemporaryFile(
                suffix=".tar", delete=False
            ) as tempArchive:
                # Add the data from the host system to the temporary archive
                tempArchive.close()
                archiveName = os.path.splitext(tempArchive.name)[0]
                shutil.make_archive(archiveName, "tar", host_path)

                # Copy the data from the temporary archive to the container
                with open(tempArchive.name, "rb") as archive:
                    container.put_archive(container_path, archive.read())

                # Remove the temporary archive
                os.unlink(tempArchive.name)

    @staticmethod
    def exec(container: Container, command: [str], capture: bool = False, **kwargs):
        """
        Executes a command in a container returned by `ContainerUtils.start_for_exec()` and streams or captures the output
        """

        # Determine if we are capturing the output or printing it
        stdoutDest = io.StringIO() if capture else sys.stdout
        stderrDest = io.StringIO() if capture else sys.stderr

        # Attempt to start the command
        details = container.client.api.exec_create(container.id, command, **kwargs)
        output = container.client.api.exec_start(details["Id"], stream=True, demux=True)

        # Stream the output
        for chunk in output:
            # Isolate the stdout and stderr chunks
            stdout, stderr = chunk

            # Capture/print the stderr data if we have any
            if stderr is not None:
                print(stderr.decode("utf-8"), end="", flush=True, file=stderrDest)

            # Capture/print the stdout data if we have any
            if stdout is not None:
                print(stdout.decode("utf-8"), end="", flush=True, file=stdoutDest)

        # Determine if the command succeeded
        capturedOutput = (
            (stdoutDest.getvalue(), stderrDest.getvalue()) if capture else None
        )
        result = container.client.api.exec_inspect(details["Id"])["ExitCode"]
        if result != 0:
            container.stop()
            raise RuntimeError(
                "Failed to run command {} in container. Process returned exit code {} with output {}.".format(
                    command,
                    result,
                    capturedOutput if capture else "printed above",
                )
            )

        # If we captured the output then return it
        return capturedOutput

    @staticmethod
    def start_for_exec(
        client: docker.DockerClient, image: str, platform: str, **kwargs
    ) -> Container:
        """
        Starts a container in a detached state using a command that will block indefinitely
        and returns the container handle. The handle can then be used to execute commands
        inside the container. The container will be removed automatically when it is stopped,
        but it will need to be stopped manually by calling `ContainerUtils.stop()`.
        """
        command = (
            ["timeout", "/t", "99999", "/nobreak"]
            if platform == "windows"
            else ["bash", "-c", "sleep infinity"]
        )
        return client.containers.run(
            image,
            command,
            stdin_open=platform == "windows",
            tty=platform == "windows",
            detach=True,
            remove=True,
            **kwargs,
        )



#!/usr/bin/env python3
import os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Comment out the call to the UE4 prereqs installer in Setup.bat
PREREQ_CALL = "start /wait Engine\\Extras\\Redist\\en-us\\UE4PrereqSetup_x64.exe"
setupScript = sys.argv[1]
verboseOutput = len(sys.argv) > 2 and sys.argv[2] == "1"
code = readFile(setupScript)
code = code.replace(
    "echo Installing prerequisites...", "echo (Skipping installation of prerequisites)"
)
code = code.replace(PREREQ_CALL, "@rem " + PREREQ_CALL)

# Also comment out the version selector call, since we don't need shell integration
SELECTOR_CALL = (
    ".\\Engine\\Binaries\\Win64\\UnrealVersionSelector-Win64-Shipping.exe /register"
)
code = code.replace(SELECTOR_CALL, "@rem " + SELECTOR_CALL)

# Add output so we can see when script execution is complete, and ensure `pause` is not called on error
code = code.replace("rem Done!", "echo Done!\r\nexit /b 0")
code = code.replace("pause", "@rem pause")
writeFile(setupScript, code)

# Print the patched code to stderr for debug purposes
if verboseOutput == True:
    print("PATCHED {}:\n\n{}".format(setupScript, code), file=sys.stderr)
else:
    print("PATCHED {}".format(setupScript), file=sys.stderr)



#!/usr/bin/env python3
import json, os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Read the build graph XML
buildXml = sys.argv[1]
code = readFile(buildXml)

# Read the UE4 version information
versionFile = sys.argv[2]
versionData = json.loads(readFile(versionFile))

# Add verbose output flags to the `BuildDerivedDataCache` command
code = code.replace(
    'Command Name="BuildDerivedDataCache" Arguments="',
    'Command Name="BuildDerivedDataCache" Arguments="-Verbose -AllowStdOutLogVerbosity ',
)

# Disable AArch64 by default (enabled since 4.24.0)
code = code.replace(
    'Property Name="DefaultWithLinuxAArch64" Value="true"',
    'Property Name="DefaultWithLinuxAArch64" Value="false"',
)
# AArch64 was renamed to Arm64 in UE-5.0, so also disable it
code = code.replace(
    'Property Name="DefaultWithLinuxArm64" Value="true"',
    'Property Name="DefaultWithLinuxArm64" Value="false"',
)

# Enable client and server targets by default in 4.23.0 onwards, except for 4.24.0 - 4.24.2 where Linux server builds fail
# (See <https://issues.unrealengine.com/issue/UE-87878> for details of the bug and its fix)
if (
    versionData["MajorVersion"] != 4
    or versionData["MinorVersion"] != 24
    or versionData["PatchVersion"] >= 3
):
    code = code.replace(
        'Option Name="WithClient" Restrict="true|false" DefaultValue="false"',
        'Option Name="WithClient" Restrict="true|false" DefaultValue="true"',
    )
    code = code.replace(
        'Option Name="WithServer" Restrict="true|false" DefaultValue="false"',
        'Option Name="WithServer" Restrict="true|false" DefaultValue="true"',
    )

# Write the modified XML back to disk
writeFile(buildXml, code)



import platform

from packaging.version import Version


class DarwinUtils(object):
    @staticmethod
    def minimumRequiredVersion() -> Version:
        """
        Returns the minimum required version of macOS, which is 10.10.3 Yosemite

        (10.10.3 is the minimum required version for Docker for Mac, as per:
        <https://store.docker.com/editions/community/docker-ce-desktop-mac>)
        """
        return Version("10.10.3")

    @staticmethod
    def systemString():
        """
        Generates a human-readable version string for the macOS host system
        """
        return "macOS {} (Kernel Version {})".format(
            platform.mac_ver()[0], platform.uname().release
        )

    @staticmethod
    def getMacOsVersion() -> Version:
        """
        Returns the version number for the macOS host system
        """
        return Version(platform.mac_ver()[0])

    @staticmethod
    def isSupportedMacOsVersion():
        """
        Verifies that the macOS host system meets our minimum version requirements
        """
        return DarwinUtils.getMacOsVersion() >= DarwinUtils.minimumRequiredVersion()



#!/usr/bin/env python3
import os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Enable the OpenGL RHI for Engine versions where it is present but deprecated
iniFile = sys.argv[1]
config = readFile(iniFile)
config = config.replace("; +TargetedRHIs=GLSL_430", "+TargetedRHIs=GLSL_430")
writeFile(iniFile, config)



import json
import platform
import random

import humanfriendly
from packaging.version import Version, InvalidVersion

from .DockerUtils import DockerUtils
from .WindowsUtils import WindowsUtils

# The default Unreal Engine git repository
DEFAULT_GIT_REPO = "https://github.com/EpicGames/UnrealEngine.git"

# The base images for Linux containers
LINUX_BASE_IMAGES = {
    "opengl": "nvidia/opengl:1.0-glvnd-devel-{ubuntu}",
    "cudagl": "nvidia/cudagl:{cuda}-devel-{ubuntu}",
}

# The default ubuntu base to use
DEFAULT_LINUX_VERSION = "ubuntu18.04"

# The default CUDA version to use when `--cuda` is specified without a value
DEFAULT_CUDA_VERSION = "11.4"

# The default memory limit (in GB) under Windows
DEFAULT_MEMORY_LIMIT = 10.0

# The Perforce changelist numbers for each supported .0 release of the Unreal Engine
UNREAL_ENGINE_RELEASE_CHANGELISTS = {
    "4.20.0": 4212847,
    "4.21.0": 4541578,
    "4.22.0": 5660361,
    "4.23.0": 8386587,
    "4.24.0": 10570792,
    "4.25.0": 13144385,
    "4.26.0": 14830424,
    "4.27.0": 17155196,
    "5.0.0": 19505902,
    "5.1.0": 23058290,
}


class VisualStudio(object):
    VS2017 = "2017"
    VS2019 = "2019"
    VS2022 = "2022"

    BuildNumbers = {
        VS2017: "15",
        VS2019: "16",
        VS2022: "17",
    }

    SupportedSince = {
        # We do not support versions older than 4.20
        VS2017: Version("4.20"),
        # Unreal Engine 4.23.1 is the first that successfully builds with Visual Studio v16.3
        # See https://github.com/EpicGames/UnrealEngine/commit/2510d4fd07a35ba5bff6ac2c7becaa6e8b7f11fa
        #
        # Unreal Engine 4.25 is the first that works with .NET SDK 4.7+
        # See https://github.com/EpicGames/UnrealEngine/commit/5256eedbdef30212ab69fdf4c09e898098959683
        VS2019: Version("4.25"),
        VS2022: Version("5.0.0"),
    }

    UnsupportedSince = {VS2017: Version("5.0")}


class ExcludedComponent(object):
    """
    The different components that we support excluding from the built images
    """

    # Engine Derived Data Cache (DDC)
    DDC = "ddc"

    # Engine debug symbols
    Debug = "debug"

    # Template projects and samples
    Templates = "templates"

    @staticmethod
    def description(component):
        """
        Returns a human-readable description of the specified component
        """
        return {
            ExcludedComponent.DDC: "Derived Data Cache (DDC)",
            ExcludedComponent.Debug: "Debug symbols",
            ExcludedComponent.Templates: "Template projects and samples",
        }.get(component, "[Unknown component]")


class BuildConfiguration(object):
    @staticmethod
    def addArguments(parser):
        """
        Registers our supported command-line arguments with the supplied argument parser
        """
        parser.add_argument(
            "release",
            nargs="?",  # aka "required = False", but that doesn't work in positionals
            help='UE4 release to build, in semver format (e.g. 4.20.0) or "custom" for a custom repo and branch (deprecated, use --ue-version instead)',
        )
        parser.add_argument(
            "--ue-version",
            default=None,
            help='UE4 release to build, in semver format (e.g. 4.20.0) or "custom" for a custom repo and branch',
        )
        parser.add_argument(
            "--linux",
            action="store_true",
            help="Build Linux container images under Windows",
        )
        parser.add_argument(
            "--rebuild",
            action="store_true",
            help="Rebuild images even if they already exist",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            help="Print `docker build` commands instead of running them",
        )
        parser.add_argument(
            "--no-minimal",
            action="store_true",
            help="Don't build the ue4-minimal image (deprecated, use --target instead)",
        )
        parser.add_argument(
            "--no-full",
            action="store_true",
            help="Don't build the ue4-full image (deprecated, use --target instead)",
        )
        parser.add_argument(
            "--no-cache", action="store_true", help="Disable Docker build cache"
        )
        parser.add_argument(
            "--target",
            action="append",
            help="Add a target to the build list. Valid targets are `build-prerequisites`, `source`, `engine`, `minimal`, `full`, and `all`. May be specified multiple times or comma-separated. Defaults to `all`.",
        )
        parser.add_argument(
            "--random-memory",
            action="store_true",
            help="Use a random memory limit for Windows containers",
        )
        parser.add_argument(
            "--exclude",
            action="append",
            default=[],
            choices=[
                ExcludedComponent.DDC,
                ExcludedComponent.Debug,
                ExcludedComponent.Templates,
            ],
            help="Exclude the specified component (can be specified multiple times to exclude multiple components)",
        )
        parser.add_argument(
            "--opt",
            action="append",
            default=[],
            help="Set an advanced configuration option (can be specified multiple times to specify multiple options)",
        )
        parser.add_argument(
            "--cuda",
            default=None,
            metavar="VERSION",
            help="Add CUDA support as well as OpenGL support when building Linux containers",
        )
        parser.add_argument(
            "--visual-studio",
            default=VisualStudio.VS2017,
            choices=VisualStudio.BuildNumbers.keys(),
            help="Specify Visual Studio Build Tools version to use for Windows containers",
        )
        parser.add_argument(
            "-username",
            default=None,
            help="Specify the username to use when cloning the git repository",
        )
        parser.add_argument(
            "-password",
            default=None,
            help="Specify access token or password to use when cloning the git repository",
        )
        parser.add_argument(
            "-repo",
            default=None,
            help='Set the custom git repository to clone when "custom" is specified as the release value',
        )
        parser.add_argument(
            "-branch",
            default=None,
            help='Set the custom branch/tag to clone when "custom" is specified as the release value',
        )
        parser.add_argument(
            "-isolation",
            default=None,
            help="Set the isolation mode to use for Windows containers (process or hyperv)",
        )
        parser.add_argument(
            "-basetag",
            default=None if platform.system() == "Windows" else DEFAULT_LINUX_VERSION,
            help="Operating system base image tag to use. For Linux this is the version of Ubuntu (default is ubuntu18.04). "
            "For Windows this is the Windows Server Core base image tag (default is the host OS version)",
        )
        parser.add_argument(
            "-suffix", default="", help="Add a suffix to the tags of the built images"
        )
        parser.add_argument(
            "-m",
            default=None,
            help="Override the default memory limit under Windows (also overrides --random-memory)",
        )
        parser.add_argument(
            "-ue4cli",
            default=None,
            help="Override the default version of ue4cli installed in the ue4-full image",
        )
        parser.add_argument(
            "-conan-ue4cli",
            default=None,
            help="Override the default version of conan-ue4cli installed in the ue4-full image",
        )
        parser.add_argument(
            "-layout",
            default=None,
            help="Copy generated Dockerfiles to the specified directory and don't build the images",
        )
        parser.add_argument(
            "--combine",
            action="store_true",
            help="Combine generated Dockerfiles into a single multi-stage build Dockerfile",
        )
        parser.add_argument(
            "--monitor",
            action="store_true",
            help="Monitor resource usage during builds (useful for debugging)",
        )
        parser.add_argument(
            "-interval",
            type=float,
            default=20.0,
            help="Sampling interval in seconds when resource monitoring has been enabled using --monitor (default is 20 seconds)",
        )
        parser.add_argument(
            "--ignore-blacklist",
            action="store_true",
            help="Run builds even on blacklisted versions of Windows (advanced use only)",
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="Enable verbose output during builds (useful for debugging)",
        )
        parser.add_argument(
            "-changelist",
            type=int,
            default=None,
            help="Set a specific changelist number in the Unreal Engine's Build.version file",
        )
        parser.add_argument(
            "--prerequisites-dockerfile",
            default=None,
            help="Specifies path to custom ue4-build-prerequisites dockerfile",
        )

    def __init__(self, parser, argv, logger):
        """
        Creates a new build configuration based on the supplied arguments object
        """

        # If the user has specified `--cuda` without a version value, treat the value as an empty string
        argv = [arg + "=" if arg == "--cuda" else arg for arg in argv]

        # Parse the supplied command-line arguments
        self.args = parser.parse_args(argv)
        self.changelist = self.args.changelist

        # Figure out what targets we have; this is needed to find out if we need --ue-version.
        using_target_specifier_old = self.args.no_minimal or self.args.no_full
        using_target_specifier_new = self.args.target is not None

        # If we specified nothing, it's the same as specifying `all`
        if not using_target_specifier_old and not using_target_specifier_new:
            self.args.target = ["all"]
        elif using_target_specifier_old and not using_target_specifier_new:
            # Convert these to the new style
            logger.warning(
                "Using deprecated `--no-*` target specifiers; recommend changing to `--target`",
                False,
            )

            # no-minimal implies no-full
            if self.args.no_minimal:
                self.args.no_full = True

            # Change into target descriptors
            self.args.target = []

            if not self.args.no_full:
                self.args.target += ["full"]

            if not self.args.no_minimal:
                self.args.target += ["minimal"]

            # disabling these was never supported
            self.args.target += ["source"]
            self.args.target += ["build-prerequisites"]

        elif using_target_specifier_new and not using_target_specifier_old:
            # these can be token-delimited, so let's just split them apart and then remerge them into one list
            split = [item.split(",") for item in self.args.target]
            self.args.target = [item for sublist in split for item in sublist]

        elif using_target_specifier_old and using_target_specifier_new:
            # uhoh
            raise RuntimeError(
                "specified both `--target` and the old `--no-*` options; please use only `--target`!"
            )

        # Now that we have our options in `self.args.target`, evaluate our dependencies
        # In a theoretical ideal world this should be code-driven; if you find yourself adding a lot more code to this, consider a redesign!
        active_targets = set(self.args.target)

        # build-prereq -> source -> engine
        # build-prereq -> source -> minimal -> full

        # We initialize these with all the options, with the intent that you should be accessing them directly and not checking for existence
        # This is to avoid typos giving false-negatives; KeyError is reliable and tells you what you did wrong
        self.buildTargets = {
            "build-prerequisites": False,
            "source": False,
            "minimal": False,
            "full": False,
        }

        for target in active_targets:
            if target != "all" and target not in self.buildTargets:
                valid_options = sorted(self.buildTargets.keys())
                raise RuntimeError(
                    f"unknown build target '{target}', valid options are: all {' '.join(valid_options)}"
                )

        if "full" in active_targets or "all" in active_targets:
            self.buildTargets["full"] = True
            active_targets.add("minimal")

        if "minimal" in active_targets or "all" in active_targets:
            self.buildTargets["minimal"] = True
            active_targets.add("source")

        if "source" in active_targets or "all" in active_targets:
            self.buildTargets["source"] = True
            active_targets.add("build-prerequisites")

        if "build-prerequisites" in active_targets or "all" in active_targets:
            self.buildTargets["build-prerequisites"] = True

        if not self.buildTargets["build-prerequisites"]:
            raise RuntimeError(
                "we're not building anything; this shouldn't even be possible, but is definitely not useful"
            )

        # See if the user specified both the old positional version option and the new ue-version option
        if self.args.release is not None and self.args.ue_version is not None:
            raise RuntimeError(
                "specified both `--ue-version` and the old positional version option; please use only `--ue-version`!"
            )

        # For the sake of a simpler pull request, we use self.args.release as the canonical place for this data.
        # If support for the old positional version option is removed, this should be fixed.
        if self.args.ue_version is not None:
            self.args.release = self.args.ue_version

        # We care about the version number only if we're building source
        if self.buildTargets["source"]:
            if self.args.release is None:
                raise RuntimeError("missing `--ue-version` when building source")

            # Determine if we are building a custom version of UE4 rather than an official release
            self.args.release = self.args.release.lower()
            if self.args.release == "custom" or self.args.release.startswith("custom:"):
                # Both a custom repository and a custom branch/tag must be specified
                if self.args.repo is None or self.args.branch is None:
                    raise RuntimeError(
                        "both a repository and branch/tag must be specified when building a custom version of the Engine"
                    )

                # Use the specified repository and branch/tag
                customName = (
                    self.args.release.split(":", 2)[1].strip()
                    if ":" in self.args.release
                    else ""
                )
                self.release = customName if len(customName) > 0 else "custom"
                self.repository = self.args.repo
                self.branch = self.args.branch
                self.custom = True

            else:
                # Validate the specified version string
                try:
                    ue4Version = Version(self.args.release)
                    if ue4Version.major not in [4, 5] or ue4Version.pre != 0:
                        raise Exception()
                    self.release = (
                        f"{ue4Version.major}.{ue4Version.minor}.{ue4Version.micro}"
                    )
                except InvalidVersion:
                    raise RuntimeError(
                        'invalid Unreal Engine release number "{}", full semver format required (e.g. "4.20.0")'.format(
                            self.args.release
                        )
                    )

                # Use the default repository and the release tag for the specified version
                self.repository = DEFAULT_GIT_REPO
                self.branch = "{}-release".format(self.release)
                self.custom = False

                # If the user specified a .0 release of the Unreal Engine and did not specify a changelist override then
                # use the official changelist number for that release to ensure consistency with Epic Games Launcher builds
                # (This is necessary because .0 releases do not include a `CompatibleChangelist` value in Build.version)
                if (
                    self.changelist is None
                    and self.release in UNREAL_ENGINE_RELEASE_CHANGELISTS
                ):
                    self.changelist = UNREAL_ENGINE_RELEASE_CHANGELISTS[self.release]
        else:
            # defaults needed by other parts of the codebase
            self.custom = False
            self.release = None
            self.repository = None
            self.branch = None

        # Store our common configuration settings
        self.containerPlatform = (
            "windows"
            if platform.system() == "Windows" and self.args.linux == False
            else "linux"
        )
        self.dryRun = self.args.dry_run
        self.rebuild = self.args.rebuild
        self.suffix = self.args.suffix
        self.platformArgs = ["--no-cache"] if self.args.no_cache == True else []
        self.excludedComponents = set(self.args.exclude)
        self.baseImage = None
        self.prereqsTag = None
        self.ignoreBlacklist = self.args.ignore_blacklist
        self.verbose = self.args.verbose
        self.layoutDir = self.args.layout
        self.combine = self.args.combine

        # If the user specified custom version strings for ue4cli and/or conan-ue4cli, process them
        self.ue4cliVersion = self._processPackageVersion("ue4cli", self.args.ue4cli)
        self.conanUe4cliVersion = self._processPackageVersion(
            "conan-ue4cli", self.args.conan_ue4cli
        )

        # Process any specified advanced configuration options (which we use directly as context values for the Jinja templating system)
        self.opts = {}
        for o in self.args.opt:
            if "=" in o:
                key, value = o.split("=", 1)
                self.opts[key.replace("-", "_")] = self._processTemplateValue(value)
            else:
                self.opts[o.replace("-", "_")] = True

        # If we are generating Dockerfiles then generate them for all images that have not been explicitly excluded
        if self.layoutDir is not None:
            self.rebuild = True

        # If we are generating Dockerfiles and combining them then set the corresponding Jinja context value
        if self.layoutDir is not None and self.combine == True:
            self.opts["combine"] = True

        # If the user requested an option that is only compatible with generated Dockerfiles then ensure `-layout` was specified
        if self.layoutDir is None and self.opts.get("source_mode", "git") != "git":
            raise RuntimeError(
                "the `-layout` flag must be used when specifying a non-default value for the `source_mode` option"
            )
        if self.layoutDir is None and self.combine == True:
            raise RuntimeError(
                "the `-layout` flag must be used when specifying the `--combine` flag"
            )

        # We care about source_mode and credential_mode only if we're building source
        if self.buildTargets["source"]:
            # Verify that the value for `source_mode` is valid if specified
            validSourceModes = ["git", "copy"]
            if self.opts.get("source_mode", "git") not in validSourceModes:
                raise RuntimeError(
                    "invalid value specified for the `source_mode` option, valid values are {}".format(
                        validSourceModes
                    )
                )

            if "credential_mode" not in self.opts:
                # On Linux, default to secrets mode that causes fewer issues with firewalls
                self.opts["credential_mode"] = (
                    "secrets" if self.containerPlatform == "linux" else "endpoint"
                )

            # Verify that the value for `credential_mode` is valid if specified
            validCredentialModes = (
                ["endpoint", "secrets"]
                if self.containerPlatform == "linux"
                else ["endpoint"]
            )
            if self.opts["credential_mode"] not in validCredentialModes:
                raise RuntimeError(
                    "invalid value specified for the `credential_mode` option, valid values are {} when building {} containers".format(
                        validCredentialModes, self.containerPlatform.title()
                    )
                )

        # Generate Jinja context values for keeping or excluding components
        self.opts["excluded_components"] = {
            "ddc": ExcludedComponent.DDC in self.excludedComponents,
            "debug": ExcludedComponent.Debug in self.excludedComponents,
            "templates": ExcludedComponent.Templates in self.excludedComponents,
        }

        if "gitdependencies_args" not in self.opts:
            self.opts["gitdependencies_args"] = (
                "--exclude=Android --exclude=Mac --exclude=Linux"
                if self.containerPlatform == "windows"
                else "--exclude=Android --exclude=Mac --exclude=Win32 --exclude=Win64"
            )

        # Warn user that they are in danger of Docker 20GB COPY bug
        # Unfortunately, we don't have a cheap way to check whether user environment is affected
        # See https://github.com/adamrehn/ue4-docker/issues/99
        if self.containerPlatform == "windows":
            warn20GiB = False
            if ExcludedComponent.Debug not in self.excludedComponents:
                logger.warning("Warning: You didn't pass --exclude debug", False)
                warn20GiB = True
            if self.release and not self.custom and Version(self.release).major >= 5:
                logger.warning("Warning: You're building Unreal Engine 5", False)
                warn20GiB = True

            if warn20GiB:
                logger.warning("Warning: You might hit Docker 20GiB COPY bug", False)
                logger.warning(
                    "Warning: Make sure that `ue4-docker diagnostics 20gig` passes",
                    False,
                )
                logger.warning(
                    "Warning: See https://github.com/adamrehn/ue4-docker/issues/99#issuecomment-1079702817 for details and workarounds",
                    False,
                )

        # If we're building Windows containers, generate our Windows-specific configuration settings
        if self.containerPlatform == "windows":
            self._generateWindowsConfig()

        # If we're building Linux containers, generate our Linux-specific configuration settings
        if self.containerPlatform == "linux":
            self._generateLinuxConfig()

        # If the user-specified suffix passed validation, prefix it with a dash
        self.suffix = "-{}".format(self.suffix) if self.suffix != "" else ""

    def describeExcludedComponents(self):
        """
        Returns a list of strings describing the components that will be excluded (if any.)
        """
        return sorted(
            [
                ExcludedComponent.description(component)
                for component in self.excludedComponents
            ]
        )

    def _generateWindowsConfig(self):
        self.visualStudio = self.args.visual_studio

        if self.release is not None and not self.custom:
            # Check whether specified Unreal Engine release is compatible with specified Visual Studio
            supportedSince = VisualStudio.SupportedSince.get(self.visualStudio, None)
            if supportedSince is not None and Version(self.release) < supportedSince:
                raise RuntimeError(
                    "specified version of Unreal Engine is too old for Visual Studio {}".format(
                        self.visualStudio
                    )
                )

            unsupportedSince = VisualStudio.UnsupportedSince.get(
                self.visualStudio, None
            )
            if (
                unsupportedSince is not None
                and Version(self.release) >= unsupportedSince
            ):
                raise RuntimeError(
                    "Visual Studio {} is too old for specified version of Unreal Engine".format(
                        self.visualStudio
                    )
                )

        self.visualStudioBuildNumber = VisualStudio.BuildNumbers[self.visualStudio]
        # See https://github.com/EpicGames/UnrealEngine/commit/72585138472785e2ee58aab9950a7260275ee2ac
        # Note: We must not pass VS2019 arg for older UE4 versions that didn't have VS2019 variable in their build graph xml.
        # Otherwise, UAT errors out with "Unknown argument: VS2019".
        if self.visualStudio != VisualStudio.VS2017:
            self.opts["buildgraph_args"] = (
                self.opts.get("buildgraph_args", "")
                + f" -set:VS{self.visualStudio}=true"
            )

        # Determine base tag for the Windows release of the host system
        self.hostBasetag = WindowsUtils.getHostBaseTag()

        # Store the tag for the base Windows Server Core image
        self.basetag = (
            self.args.basetag if self.args.basetag is not None else self.hostBasetag
        )

        if self.basetag is None:
            raise RuntimeError(
                "unable to determine Windows Server Core base image tag from host system. Specify it explicitly using -basetag command-line flag"
            )

        self.baseImage = "mcr.microsoft.com/windows/servercore:" + self.basetag
        self.dllSrcImage = WindowsUtils.getDllSrcImage(self.basetag)
        self.prereqsTag = self.basetag + "-vs" + self.visualStudio

        # If the user has explicitly specified an isolation mode then use it, otherwise auto-detect
        if self.args.isolation is not None:
            self.isolation = self.args.isolation
        else:
            # If we are able to use process isolation mode then use it, otherwise use Hyper-V isolation mode
            differentKernels = self.basetag != self.hostBasetag
            dockerSupportsProcess = Version(
                DockerUtils.version()["Version"]
            ) >= Version("18.09.0")
            if not differentKernels and dockerSupportsProcess:
                self.isolation = "process"
            else:
                self.isolation = "hyperv"

        # Set the isolation mode Docker flag
        self.platformArgs.append("--isolation=" + self.isolation)

        # If the user has explicitly specified a memory limit then use it, otherwise auto-detect
        self.memLimit = None
        if self.args.m is not None:
            try:
                self.memLimit = humanfriendly.parse_size(self.args.m) / (
                    1000 * 1000 * 1000
                )
            except:
                raise RuntimeError('invalid memory limit "{}"'.format(self.args.m))
        else:
            # Only specify a memory limit when using Hyper-V isolation mode, in order to override the 1GB default limit
            # (Process isolation mode does not impose any memory limits by default)
            if self.isolation == "hyperv":
                self.memLimit = (
                    DEFAULT_MEMORY_LIMIT
                    if self.args.random_memory == False
                    else random.uniform(
                        DEFAULT_MEMORY_LIMIT, DEFAULT_MEMORY_LIMIT + 2.0
                    )
                )

        # Set the memory limit Docker flag
        if self.memLimit is not None:
            self.platformArgs.extend(["-m", "{:.2f}GB".format(self.memLimit)])

    def _generateLinuxConfig(self):
        # Verify that any user-specified tag suffix does not collide with our base tags
        if self.suffix.startswith("opengl") or self.suffix.startswith("cudagl"):
            raise RuntimeError('tag suffix cannot begin with "opengl" or "cudagl".')

        # Determine if we are building CUDA-enabled container images
        self.cuda = None
        if self.args.cuda is not None:
            # Verify that the specified CUDA version is valid
            self.cuda = self.args.cuda if self.args.cuda != "" else DEFAULT_CUDA_VERSION
            # Use the appropriate base image for the specified CUDA version
            self.baseImage = LINUX_BASE_IMAGES["cudagl"]
            self.prereqsTag = "cudagl{cuda}-{ubuntu}"
        else:
            self.baseImage = LINUX_BASE_IMAGES["opengl"]
            self.prereqsTag = "opengl-{ubuntu}"

        self.baseImage = self.baseImage.format(
            cuda=self.args.cuda, ubuntu=self.args.basetag
        )
        self.prereqsTag = self.prereqsTag.format(
            cuda=self.args.cuda, ubuntu=self.args.basetag
        )

    def _processPackageVersion(self, package, version):
        # Leave the version value unmodified if a blank version was specified or a fully-qualified version was specified
        # (e.g. package==X.X.X, package>=X.X.X, git+https://url/for/package/repo.git, etc.)
        if version is None or "/" in version or version.lower().startswith(package):
            return version

        # If a version specifier (e.g. ==X.X.X, >=X.X.X, etc.) was specified, prefix it with the package name
        if "=" in version:
            return package + version

        # If a raw version number was specified, prefix the package name and a strict equality specifier
        return "{}=={}".format(package, version)

    def _processTemplateValue(self, value):
        # If the value is a boolean (either raw or represented by zero or one) then parse it
        if value.lower() in ["true", "1"]:
            return True
        elif value.lower() in ["false", "0"]:
            return False

        # If the value is a JSON object or array then attempt to parse it
        if (value.startswith("{") and value.endswith("}")) or (
            value.startswith("[") and value.endswith("]")
        ):
            try:
                return json.loads(value)
            except:
                print(
                    'Warning: could not parse option value "{}" as JSON, treating value as a string'.format(
                        value
                    )
                )

        # Treat all other values as strings
        return value



from .main import main
import os, sys

if __name__ == "__main__":
    # Rewrite sys.argv[0] so our help prompts display the correct base command
    interpreter = sys.executable if sys.executable not in [None, ""] else "python3"
    sys.argv[0] = "{} -m ue4docker".format(os.path.basename(interpreter))
    main()



class FilesystemUtils(object):
    @staticmethod
    def readFile(filename):
        """
        Reads data from a file
        """
        with open(filename, "rb") as f:
            return f.read().decode("utf-8")

    @staticmethod
    def writeFile(filename, data):
        """
        Writes data to a file
        """
        with open(filename, "wb") as f:
            f.write(data.encode("utf-8"))



import argparse, getpass, humanfriendly, json, os, shutil, sys, tempfile, time
from .infrastructure import *
from .version import __version__
from os.path import join


def _getCredential(args, name, envVar, promptFunc):
    # Check if the credential was specified via the command-line
    if getattr(args, name, None) is not None:
        print("Using {} specified via `-{}` command-line argument.".format(name, name))
        return getattr(args, name)

    # Check if the credential was specified via an environment variable
    if envVar in os.environ:
        print("Using {} specified via {} environment variable.".format(name, envVar))
        return os.environ[envVar]

    # Fall back to prompting the user for the value
    return promptFunc()


def _getUsername(args):
    return _getCredential(
        args, "username", "UE4DOCKER_USERNAME", lambda: input("Username: ")
    )


def _getPassword(args):
    return _getCredential(
        args,
        "password",
        "UE4DOCKER_PASSWORD",
        lambda: getpass.getpass("Access token or password: "),
    )


def build():
    # Create our logger to generate coloured output on stderr
    logger = Logger(prefix="[{} build] ".format(sys.argv[0]))

    # Register our supported command-line arguments
    parser = argparse.ArgumentParser(prog="{} build".format(sys.argv[0]))
    BuildConfiguration.addArguments(parser)

    # If no command-line arguments were supplied, display the help message and exit
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    # Parse the supplied command-line arguments
    try:
        config = BuildConfiguration(parser, sys.argv[1:], logger)
    except RuntimeError as e:
        logger.error("Error: {}".format(e))
        sys.exit(1)

    # Verify that Docker is installed
    if DockerUtils.installed() == False:
        logger.error(
            "Error: could not detect Docker version. Please ensure Docker is installed."
        )
        sys.exit(1)

    # Verify that we aren't trying to build Windows containers under Windows 10 when in Linux container mode (or vice versa)
    # (Note that we don't bother performing this check when we're just copying Dockerfiles to an output directory)
    if config.layoutDir is None:
        dockerPlatform = DockerUtils.info()["OSType"].lower()
        if config.containerPlatform == "windows" and dockerPlatform == "linux":
            logger.error(
                "Error: cannot build Windows containers when Docker Desktop is in Linux container",
                False,
            )
            logger.error(
                "mode. Use the --linux flag if you want to build Linux containers instead.",
                False,
            )
            sys.exit(1)
        elif config.containerPlatform == "linux" and dockerPlatform == "windows":
            logger.error(
                "Error: cannot build Linux containers when Docker Desktop is in Windows container",
                False,
            )
            logger.error(
                "mode. Remove the --linux flag if you want to build Windows containers instead.",
                False,
            )
            sys.exit(1)

    # Create an auto-deleting temporary directory to hold our build context
    with tempfile.TemporaryDirectory() as tempDir:
        contextOrig = join(os.path.dirname(os.path.abspath(__file__)), "dockerfiles")

        # Create the builder instance to build the Docker images
        builder = ImageBuilder(
            join(tempDir, "dockerfiles"),
            config.containerPlatform,
            logger,
            config.rebuild,
            config.dryRun,
            config.layoutDir,
            config.opts,
            config.combine,
        )

        # Resolve our main set of tags for the generated images; this is used only for Source and downstream
        if config.buildTargets["source"]:
            mainTags = [
                "{}{}-{}".format(config.release, config.suffix, config.prereqsTag),
                config.release + config.suffix,
            ]

        # Print the command-line invocation that triggered this build, masking any supplied passwords
        args = [
            "*******"
            if config.args.password is not None and arg == config.args.password
            else arg
            for arg in sys.argv
        ]
        logger.info("COMMAND-LINE INVOCATION:", False)
        logger.info(str(args), False)

        # Print the details of the Unreal Engine version being built
        logger.info("UNREAL ENGINE VERSION SETTINGS:")
        logger.info(
            "Custom build:  {}".format("Yes" if config.custom == True else "No"), False
        )
        if config.custom == True:
            logger.info("Custom name:   " + config.release, False)
        elif config.release is not None:
            logger.info("Release:       " + config.release, False)
        if config.repository is not None:
            logger.info("Repository:    " + config.repository, False)
            logger.info("Branch/tag:    " + config.branch + "\n", False)

        # Determine if we are using a custom version for ue4cli or conan-ue4cli
        if config.ue4cliVersion is not None or config.conanUe4cliVersion is not None:
            logger.info("CUSTOM PACKAGE VERSIONS:", False)
            logger.info(
                "ue4cli:        {}".format(
                    config.ue4cliVersion
                    if config.ue4cliVersion is not None
                    else "default"
                ),
                False,
            )
            logger.info(
                "conan-ue4cli:  {}\n".format(
                    config.conanUe4cliVersion
                    if config.conanUe4cliVersion is not None
                    else "default"
                ),
                False,
            )

        # Report any advanced configuration options that were specified
        if len(config.opts) > 0:
            logger.info("ADVANCED CONFIGURATION OPTIONS:", False)
            for key, value in sorted(config.opts.items()):
                logger.info("{}: {}".format(key, json.dumps(value)), False)
            print("", file=sys.stderr, flush=True)

        # Determine if we are building Windows or Linux containers
        if config.containerPlatform == "windows":
            # Provide the user with feedback so they are aware of the Windows-specific values being used
            logger.info("WINDOWS CONTAINER SETTINGS", False)
            logger.info(
                "Isolation mode:               {}".format(config.isolation), False
            )
            logger.info(
                "Base OS image:                {}".format(config.baseImage), False
            )
            logger.info(
                "Dll source image:             {}".format(config.dllSrcImage), False
            )
            logger.info(
                "Host OS:                      {}".format(WindowsUtils.systemString()),
                False,
            )
            logger.info(
                "Memory limit:                 {}".format(
                    "No limit"
                    if config.memLimit is None
                    else "{:.2f}GB".format(config.memLimit)
                ),
                False,
            )
            logger.info(
                "Detected max image size:      {:.0f}GB".format(DockerUtils.maxsize()),
                False,
            )
            logger.info(
                "Visual Studio:                {}".format(config.visualStudio), False
            )

            # Verify that the host OS is not a release that is blacklisted due to critical bugs
            if (
                config.ignoreBlacklist == False
                and WindowsUtils.isBlacklistedWindowsHost() == True
            ):
                logger.error(
                    "Error: detected blacklisted host OS version: {}".format(
                        WindowsUtils.systemString()
                    ),
                    False,
                )
                logger.error("", False)
                logger.error(
                    "This version of Windows contains one or more critical bugs that",
                    False,
                )
                logger.error(
                    "render it incapable of successfully building UE4 container images.",
                    False,
                )
                logger.error(
                    "You will need to use an older or newer version of Windows.", False
                )
                logger.error("", False)
                logger.error("For more information, see:", False)
                logger.error(
                    "https://unrealcontainers.com/docs/concepts/windows-containers",
                    False,
                )
                sys.exit(1)

            # Verify that the user is not attempting to build images with a newer kernel version than the host OS
            newer_check = WindowsUtils.isNewerBaseTag(
                config.hostBasetag, config.basetag
            )
            if newer_check:
                logger.error(
                    "Error: cannot build container images with a newer kernel version than that of the host OS!"
                )
                sys.exit(1)
            elif newer_check is None:
                logger.warning(
                    "Warning: unable to determine whether host system is new enough to use specified base tag"
                )

            # Ensure the Docker daemon is configured correctly
            requiredLimit = WindowsUtils.requiredSizeLimit()
            if DockerUtils.maxsize() < requiredLimit and config.buildTargets["source"]:
                logger.error("SETUP REQUIRED:")
                logger.error(
                    "The max image size for Windows containers must be set to at least {}GB.".format(
                        requiredLimit
                    )
                )
                logger.error(
                    "See the Microsoft documentation for configuration instructions:"
                )
                logger.error(
                    "https://docs.microsoft.com/en-us/virtualization/windowscontainers/manage-containers/container-storage#storage-limits"
                )
                logger.error(
                    "Under Windows Server, the command `{} setup` can be used to automatically configure the system.".format(
                        sys.argv[0]
                    )
                )
                sys.exit(1)

        elif config.containerPlatform == "linux":
            logger.info("LINUX CONTAINER SETTINGS", False)
            logger.info(
                "Base OS image: {}\n".format(config.baseImage),
                False,
            )

        # Report which Engine components are being excluded (if any)
        logger.info("GENERAL SETTINGS", False)
        logger.info(
            "Build targets: {}".format(
                " ".join(
                    sorted(
                        [
                            target
                            for target, enabled in config.buildTargets.items()
                            if enabled
                        ]
                    )
                )
            ),
            False,
        )
        logger.info(
            "Changelist override: {}".format(
                config.changelist
                if config.changelist is not None
                else "(None specified)"
            ),
            False,
        )
        if len(config.excludedComponents) > 0:
            logger.info("Excluding the following Engine components:", False)
            for component in config.describeExcludedComponents():
                logger.info("- {}".format(component), False)
        else:
            logger.info("Not excluding any Engine components.", False)

        # Determine if we need to prompt for credentials
        if config.dryRun == True:
            # Don't bother prompting the user for any credentials during a dry run
            logger.info(
                "Performing a dry run, `docker build` commands will be printed and not executed.",
                False,
            )
            username = ""
            password = ""

        elif config.layoutDir is not None:
            # Don't bother prompting the user for any credentials when we're just copying the Dockerfiles to a directory
            logger.info(
                "Copying generated Dockerfiles to: {}".format(config.layoutDir), False
            )
            username = ""
            password = ""

        elif (
            not config.buildTargets["source"]
            or builder.willBuild("ue4-source", mainTags) == False
        ):
            # Don't bother prompting the user for any credentials if we're not building the ue4-source image
            logger.info(
                "Not building the ue4-source image, no Git credentials required.", False
            )
            username = ""
            password = ""

        else:
            # Retrieve the Git username and password from the user when building the ue4-source image
            print(
                "\nRetrieving the Git credentials that will be used to clone the UE4 repo"
            )
            username = _getUsername(config.args)
            password = _getPassword(config.args)
            print()

        # If resource monitoring has been enabled, start the resource monitoring background thread
        resourceMonitor = ResourceMonitor(logger, config.args.interval)
        if config.args.monitor == True:
            resourceMonitor.start()

        # Prep for endpoint cleanup, if necessary
        endpoint = None

        try:
            # Keep track of our starting time
            startTime = time.time()

            # If we're copying Dockerfiles to an output directory then make sure it exists and is empty
            if config.layoutDir is not None:
                if os.path.exists(config.layoutDir):
                    shutil.rmtree(config.layoutDir)
                os.makedirs(config.layoutDir)

            # Keep track of the images we've built
            builtImages = []

            commonArgs = [
                "--build-arg",
                "NAMESPACE={}".format(GlobalConfiguration.getTagNamespace()),
            ]

            # Build the UE4 build prerequisites image
            if config.buildTargets["build-prerequisites"]:
                # Compute the build options for the UE4 build prerequisites image
                # (This is the only image that does not use any user-supplied tag suffix, since the tag always reflects any customisations)
                prereqsArgs = ["--build-arg", "BASEIMAGE=" + config.baseImage]
                if config.containerPlatform == "windows":
                    prereqsArgs = prereqsArgs + [
                        "--build-arg",
                        "DLLSRCIMAGE=" + config.dllSrcImage,
                        "--build-arg",
                        "VISUAL_STUDIO_BUILD_NUMBER=" + config.visualStudioBuildNumber,
                    ]

                custom_prerequisites_dockerfile = config.args.prerequisites_dockerfile
                if custom_prerequisites_dockerfile is not None:
                    builder.build_builtin_image(
                        "ue4-base-build-prerequisites",
                        [config.prereqsTag],
                        commonArgs + config.platformArgs + prereqsArgs,
                        builtin_name="ue4-build-prerequisites",
                    )
                    builtImages.append("ue4-base-build-prerequisites")
                else:
                    builder.build_builtin_image(
                        "ue4-build-prerequisites",
                        [config.prereqsTag],
                        commonArgs + config.platformArgs + prereqsArgs,
                    )

                prereqConsumerArgs = [
                    "--build-arg",
                    "PREREQS_TAG={}".format(config.prereqsTag),
                ]

                if custom_prerequisites_dockerfile is not None:
                    builder.build(
                        "ue4-build-prerequisites",
                        [config.prereqsTag],
                        commonArgs + config.platformArgs + prereqConsumerArgs,
                        dockerfile_template=custom_prerequisites_dockerfile,
                        context_dir=os.path.dirname(custom_prerequisites_dockerfile),
                    )

                builtImages.append("ue4-build-prerequisites")
            else:
                logger.info("Skipping ue4-build-prerequisities image build.")

            # Build the UE4 source image
            if config.buildTargets["source"]:
                # Start the HTTP credential endpoint as a child process and wait for it to start
                if config.opts["credential_mode"] == "endpoint":
                    endpoint = CredentialEndpoint(username, password)
                    endpoint.start()

                # If we're using build secrets then pass the Git username and password to the UE4 source image as secrets
                secrets = {}
                if config.opts["credential_mode"] == "secrets":
                    secrets = {"username": username, "password": password}
                credentialArgs = [] if len(secrets) > 0 else endpoint.args()

                ue4SourceArgs = prereqConsumerArgs + [
                    "--build-arg",
                    "GIT_REPO={}".format(config.repository),
                    "--build-arg",
                    "GIT_BRANCH={}".format(config.branch),
                    "--build-arg",
                    "VERBOSE_OUTPUT={}".format("1" if config.verbose == True else "0"),
                ]
                builder.build_builtin_image(
                    "ue4-source",
                    mainTags,
                    commonArgs + config.platformArgs + ue4SourceArgs + credentialArgs,
                    secrets=secrets,
                )
                builtImages.append("ue4-source")
            else:
                logger.info("Skipping ue4-source image build.")

            if config.buildTargets["minimal"]:
                ue4BuildArgs = prereqConsumerArgs + [
                    "--build-arg",
                    "TAG={}".format(mainTags[1]),
                ]

            # Build the minimal UE4 CI image, unless requested otherwise by the user
            if config.buildTargets["minimal"]:
                minimalArgs = (
                    ["--build-arg", "CHANGELIST={}".format(config.changelist)]
                    if config.changelist is not None
                    else []
                )

                builder.build_builtin_image(
                    "ue4-minimal",
                    mainTags,
                    commonArgs + config.platformArgs + ue4BuildArgs + minimalArgs,
                )
                builtImages.append("ue4-minimal")
            else:
                logger.info("Skipping ue4-minimal image build.")

            # Build the full UE4 CI image, unless requested otherwise by the user
            if config.buildTargets["full"]:
                # If custom version strings were specified for ue4cli and/or conan-ue4cli, use them
                infrastructureFlags = []
                if config.ue4cliVersion is not None:
                    infrastructureFlags.extend(
                        [
                            "--build-arg",
                            "UE4CLI_VERSION={}".format(config.ue4cliVersion),
                        ]
                    )
                if config.conanUe4cliVersion is not None:
                    infrastructureFlags.extend(
                        [
                            "--build-arg",
                            "CONAN_UE4CLI_VERSION={}".format(config.conanUe4cliVersion),
                        ]
                    )

                # Build the image
                builder.build_builtin_image(
                    "ue4-full",
                    mainTags,
                    commonArgs
                    + config.platformArgs
                    + ue4BuildArgs
                    + infrastructureFlags,
                )
                builtImages.append("ue4-full")
            else:
                logger.info("Skipping ue4-full image build.")

            # If we are generating Dockerfiles then include information about the options used to generate them
            if config.layoutDir is not None:
                # Determine whether we generated a single combined Dockerfile or a set of Dockerfiles
                if config.combine == True:
                    # Generate a comment to place at the top of the single combined Dockerfile
                    lines = [
                        "This file was generated by ue4-docker version {} with the following options:".format(
                            __version__
                        ),
                        "",
                    ]
                    lines.extend(
                        [
                            "- {}: {}".format(key, json.dumps(value))
                            for key, value in sorted(config.opts.items())
                        ]
                    )
                    lines.extend(
                        [
                            "",
                            "This Dockerfile combines the steps for the following images:",
                            "",
                        ]
                    )
                    lines.extend(["- {}".format(image) for image in builtImages])
                    comment = "\n".join(["# {}".format(line) for line in lines])

                    # Inject the comment at the top of the Dockerfile, being sure to place it after any `escape` parser directive
                    dockerfile = join(config.layoutDir, "combined", "Dockerfile")
                    dockerfileContents = FilesystemUtils.readFile(dockerfile)
                    if dockerfileContents.startswith("# escape"):
                        newline = dockerfileContents.index("\n")
                        dockerfileContents = (
                            dockerfileContents[0 : newline + 1]
                            + "\n"
                            + comment
                            + "\n\n"
                            + dockerfileContents[newline + 1 :]
                        )
                    else:
                        dockerfileContents = comment + "\n\n" + dockerfileContents
                    FilesystemUtils.writeFile(dockerfile, dockerfileContents)

                else:
                    # Create a JSON file to accompany the set of generated Dockerfiles
                    FilesystemUtils.writeFile(
                        join(config.layoutDir, "generated.json"),
                        json.dumps(
                            {
                                "version": __version__,
                                "images": builtImages,
                                "opts": config.opts,
                            },
                            indent=4,
                            sort_keys=True,
                        ),
                    )

            # Report the total execution time
            endTime = time.time()
            logger.action(
                "Total execution time: {}".format(
                    humanfriendly.format_timespan(endTime - startTime)
                )
            )

            # Stop the resource monitoring background thread if it is running
            resourceMonitor.stop()

            # Stop the HTTP server
            if endpoint is not None:
                endpoint.stop()

        except (Exception, KeyboardInterrupt) as e:
            # One of the images failed to build
            logger.error("Error: {}".format(e))
            resourceMonitor.stop()
            if endpoint is not None:
                endpoint.stop()
            sys.exit(1)



import datetime, humanfriendly, os, psutil, shutil, threading, time
from .DockerUtils import DockerUtils


class ResourceMonitor(threading.Thread):
    def __init__(self, logger, interval):
        """
        Creates a resource monitor with the specified configuration
        """
        super().__init__()
        self._logger = logger
        self._interval = interval
        self._lock = threading.Lock()
        self._shouldStop = False

    def stop(self):
        """
        Stops the resource monitor thread
        """

        # Set the flag to instruct the resource monitor loop to stop
        with self._lock:
            self._shouldStop = True

        # Wait for the resource monitor thread to complete
        if self.is_alive() == True:
            self.join()

    def run(self):
        """
        The resource monitor loop itself
        """

        # Determine which filesystem the Docker daemon uses for storing its data directory
        dockerInfo = DockerUtils.info()
        rootDir = dockerInfo["DockerRootDir"]

        # If we cannot access the Docker data directory (e.g. when the daemon is in a Moby VM), don't report disk space
        reportDisk = os.path.exists(rootDir)

        # Sample the CPU usage using an interval of 1 second the first time to prime the system
        # (See: <https://psutil.readthedocs.io/en/latest/#psutil.cpu_percent>)
        psutil.cpu_percent(1.0)

        # Loop until asked to stop
        while True:
            # Check that the thread has not been asked to stop
            with self._lock:
                if self._shouldStop == True:
                    return

            # Format the timestamp for the current time in ISO 8601 format (albeit without the "T" separator)
            isoTime = datetime.datetime.now().replace(microsecond=0).isoformat(" ")

            # We format data sizes using binary units (KiB, MiB, GiB, etc.)
            formatSize = lambda size: humanfriendly.format_size(
                size, binary=True, keep_width=True
            )

            # Format the current quantity of available disk space on the Docker data directory's filesystem
            diskSpace = (
                formatSize(shutil.disk_usage(rootDir).free)
                if reportDisk == True
                else "Unknown"
            )

            # Format the current quantity of available system memory
            physicalMemory = formatSize(psutil.virtual_memory().free)
            virtualMemory = formatSize(psutil.swap_memory().free)

            # Format the current CPU usage levels
            cpu = psutil.cpu_percent()

            # Report the current levels of our available resources
            self._logger.info(
                "[{}] [Available disk: {}] [Available memory: {} physical, {} virtual] [CPU usage: {:.2f}%]".format(
                    isoTime, diskSpace, physicalMemory, virtualMemory, cpu
                ),
                False,
            )

            # Sleep until the next sampling interval
            time.sleep(self._interval)



from ..infrastructure import DockerUtils, WindowsUtils
from .base import DiagnosticBase
import argparse, platform


class diagnosticNetwork(DiagnosticBase):
    # The tag we use for built images
    IMAGE_TAG = "adamrehn/ue4-docker/diagnostics:network"

    def __init__(self):
        # Setup our argument parser so we can use its help message output in our description text
        self._parser = argparse.ArgumentParser(prog="ue4-docker diagnostics network")
        self._parser.add_argument(
            "--linux",
            action="store_true",
            help="Use Linux containers under Windows hosts (useful when testing Docker Desktop or LCOW support)",
        )
        self._parser.add_argument(
            "--isolation",
            default=None,
            choices=["hyperv", "process"],
            help="Override the default isolation mode when testing Windows containers",
        )
        self._parser.add_argument(
            "-basetag",
            default=None,
            choices=WindowsUtils.getKnownBaseTags(),
            help="Override the default base image tag when testing Windows containers",
        )

    def getName(self):
        """
        Returns the human-readable name of the diagnostic
        """
        return "Check that containers can access the internet correctly"

    def getDescription(self):
        """
        Returns a description of what the diagnostic does
        """
        return "\n".join(
            [
                "This diagnostic determines if running containers are able to access the internet,",
                "resolve DNS entries, and download remote files.",
                "",
                "This is primarily useful in troubleshooting network connectivity and proxy issues.",
            ]
        )

    def getPrefix(self):
        """
        Returns the short name of the diagnostic for use in log output
        """
        return "network"

    def run(self, logger, args=[]):
        """
        Runs the diagnostic
        """

        # Parse our supplied arguments
        args = self._parser.parse_args(args)

        # Determine which image platform we will build the Dockerfile for (default is the host platform unless overridden)
        containerPlatform = (
            "linux"
            if args.linux == True or platform.system().lower() != "windows"
            else "windows"
        )

        # Verify that the user isn't trying to test Windows containers under Windows 10 when in Linux container mode (or vice versa)
        try:
            self._checkPlatformMistmatch(logger, containerPlatform)
        except RuntimeError:
            return False

        # Set our build arguments when testing Windows containers
        buildArgs = (
            self._generateWindowsBuildArgs(logger, args.basetag, args.isolation)
            if containerPlatform == "windows"
            else []
        )

        # Attempt to build the Dockerfile
        logger.action(
            "[network] Attempting to build an image that accesses network resources...",
            False,
        )
        built = self._buildDockerfile(
            logger, containerPlatform, diagnosticNetwork.IMAGE_TAG, buildArgs
        )

        # Inform the user of the outcome of the diagnostic
        if built == True:
            logger.action(
                "[network] Diagnostic succeeded! Running containers can access network resources without any issues.\n"
            )
        else:
            logger.error(
                "[network] Diagnostic failed! Running containers cannot access network resources. See the docs for troubleshooting tips:",
                True,
            )
            logger.error(
                "[network] https://docs.adamrehn.com/ue4-docker/building-images/troubleshooting-build-issues#building-the-ue4-build-prerequisites-image-fails-with-a-network-related-error\n",
                False,
            )

        return built



import tempfile

from docker.models.containers import Container

from ..infrastructure import DockerUtils, SubprocessUtils
import json, os, platform, shutil, subprocess, sys


def exportInstalledBuild(image, destination, extraArgs):
    # Verify that the destination directory does not already exist
    if os.path.exists(destination) == True:
        print("Error: the destination directory already exists.", file=sys.stderr)
        sys.exit(1)

    # Create a container from which we will copy files
    container = DockerUtils.create(image)

    exit_code = 1
    try:
        exit_code = doExportInstalledBuild(container, destination, extraArgs)
    except Exception as e:
        print("Error: failed to export Installed Build.", file=sys.stderr)
        raise e
    finally:
        # Remove the container, irrespective of whether or not the export succeeded
        container.remove()

        sys.exit(exit_code)


def doExportInstalledBuild(container: Container, destination: str, extraArgs) -> int:
    if platform.system() == "Windows":
        engineRoot = "C:/UnrealEngine"
    else:
        engineRoot = "/home/ue4/UnrealEngine"

    with tempfile.TemporaryDirectory() as tmpdir:
        versionFilePath = os.path.join(tmpdir, "Build.version")
        # Verify that the Installed Build in the specified image is at least 4.21.0
        subprocess.run(
            [
                "docker",
                "cp",
                f"{container.name}:{engineRoot}/Engine/Build/Build.version",
                versionFilePath,
            ],
            check=True,
        )
        try:
            with open(versionFilePath, "r") as versionFile:
                version = json.load(versionFile)
                if version["MajorVersion"] == 4 and version["MinorVersion"] < 21:
                    raise Exception()
        except:
            print(
                "Error: Installed Builds can only be exported for Unreal Engine 4.21.0 and newer.",
                file=sys.stderr,
            )
            return 1

    # Attempt to perform the export
    print("Exporting to {}...".format(destination))
    subprocess.run(
        ["docker", "cp", f"{container.name}:{engineRoot}", destination], check=True
    )

    # If the export succeeded, regenerate the linker symlinks on the host system
    if platform.system() != "Windows":
        print("Performing linker symlink fixup...")
        subprocess.run(
            [
                sys.executable,
                os.path.join(
                    os.path.dirname(os.path.dirname(__file__)),
                    "dockerfiles",
                    "ue4-source",
                    "linux",
                    "linker-fixup.py",
                ),
                os.path.join(
                    destination,
                    "Engine/Extras/ThirdPartyNotUE/SDKs/HostLinux/Linux_x64",
                ),
                shutil.which("ld"),
            ],
            check=True,
        )



import multiprocessing, secrets, time, urllib.parse
from .NetworkUtils import NetworkUtils
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs
from functools import partial


class CredentialRequestHandler(BaseHTTPRequestHandler):
    def __init__(self, username: str, password: str, token: str, *args, **kwargs):
        self.username = username
        self.password = password
        self.token = token
        super().__init__(*args, **kwargs)

    def log_request(self, code: str = "-", size: str = "-") -> None:
        # We do not want to log each and every incoming request
        pass

    def do_POST(self):
        query_components = parse_qs(urlparse(self.path).query)

        self.send_response(200)
        self.end_headers()

        if "token" in query_components and query_components["token"][0] == self.token:
            content_length = int(self.headers["Content-Length"])
            prompt = self.rfile.read(content_length).decode("utf-8")

            response = self.password if "Password for" in prompt else self.username
            self.wfile.write(response.encode("utf-8"))


class CredentialEndpoint(object):
    def __init__(self, username: str, password: str):
        """
        Creates an endpoint manager for the supplied credentials
        """

        # Make sure neither our username or password are blank, since that can cause `git clone` to hang indefinitely
        self.username = username if username is not None and len(username) > 0 else " "
        self.password = password if password is not None and len(password) > 0 else " "
        self.endpoint = None

        # Generate a security token to require when requesting credentials
        self.token = secrets.token_hex(16)

    def args(self) -> [str]:
        """
        Returns the Docker build arguments for creating containers that require Git credentials
        """

        # Resolve the IP address for the host system
        hostAddress = NetworkUtils.hostIP()

        # Provide the host address and security token to the container
        return [
            "--build-arg",
            "HOST_ADDRESS_ARG=" + urllib.parse.quote_plus(hostAddress),
            "--build-arg",
            "HOST_TOKEN_ARG=" + urllib.parse.quote_plus(self.token),
        ]

    def start(self) -> None:
        """
        Starts the HTTP endpoint as a child process
        """

        # Create a child process to run the credential endpoint
        self.endpoint = multiprocessing.Process(
            target=CredentialEndpoint._endpoint,
            args=(self.username, self.password, self.token),
        )

        # Spawn the child process and give the endpoint time to start
        self.endpoint.start()
        time.sleep(2)

        # Verify that the endpoint started correctly
        if not self.endpoint.is_alive():
            raise RuntimeError("failed to start the credential endpoint")

    def stop(self) -> None:
        """
        Stops the HTTP endpoint child process
        """
        self.endpoint.terminate()
        self.endpoint.join()

    @staticmethod
    def _endpoint(username: str, password: str, token: str) -> None:
        """
        Implements a HTTP endpoint to provide Git credentials to Docker containers
        """
        handler = partial(CredentialRequestHandler, username, password, token)

        server = HTTPServer(("0.0.0.0", 9876), RequestHandlerClass=handler)
        server.serve_forever()



import humanfriendly, platform, psutil, shutil, sys
from .version import __version__
from .infrastructure import *


def _osName(dockerInfo):
    if platform.system() == "Windows":
        return WindowsUtils.systemString()
    elif platform.system() == "Darwin":
        return DarwinUtils.systemString()
    else:
        return "Linux ({}, {})".format(
            dockerInfo["OperatingSystem"], dockerInfo["KernelVersion"]
        )


def _formatSize(size):
    return humanfriendly.format_size(size, binary=True)


def info():
    # Verify that Docker is installed
    if DockerUtils.installed() == False:
        print(
            "Error: could not detect Docker version. Please ensure Docker is installed.",
            file=sys.stderr,
        )
        sys.exit(1)

    # Gather our information about the Docker daemon
    dockerInfo = DockerUtils.info()
    nvidiaDocker = platform.system() == "Linux" and "nvidia" in dockerInfo["Runtimes"]
    maxSize = DockerUtils.maxsize()
    rootDir = dockerInfo["DockerRootDir"]

    # If we are communicating with a Linux Docker daemon under Windows or macOS then we can't query the available disk space
    canQueryDisk = dockerInfo["OSType"].lower() == platform.system().lower()

    # Gather our information about the host system
    diskSpace = (
        _formatSize(shutil.disk_usage(rootDir).free)
        if canQueryDisk == True
        else "Unknown (typically means the Docker daemon is running in a Moby VM, e.g. Docker Desktop)"
    )
    memPhysical = psutil.virtual_memory().total
    memVirtual = psutil.swap_memory().total
    cpuPhysical = psutil.cpu_count(False)
    cpuLogical = psutil.cpu_count()
    cpuModel = platform.processor()

    # Attempt to query PyPI to determine the latest version of ue4-docker
    # (We ignore any errors here to ensure the `ue4-docker info` command still works without network access)
    try:
        latestVersion = GlobalConfiguration.getLatestVersion()
    except:
        latestVersion = None

    # Prepare our report items
    items = [
        (
            "ue4-docker version",
            "{}{}".format(
                __version__,
                ""
                if latestVersion is None
                else " (latest available version is {})".format(latestVersion),
            ),
        ),
        ("Operating system", _osName(dockerInfo)),
        ("Docker daemon version", dockerInfo["ServerVersion"]),
        ("NVIDIA Docker supported", "Yes" if nvidiaDocker == True else "No"),
        (
            "Maximum image size",
            "{:.0f}GB".format(maxSize) if maxSize != -1 else "No limit detected",
        ),
        ("Available disk space", diskSpace),
        (
            "Total system memory",
            "{} physical, {} virtual".format(
                _formatSize(memPhysical), _formatSize(memVirtual)
            ),
        ),
        (
            "CPU",
            "{} physical, {} logical ({})".format(cpuPhysical, cpuLogical, cpuModel),
        ),
    ]

    # Determine the longest item name so we can format our list in nice columns
    longestName = max([len(i[0]) for i in items])
    minSpaces = 4

    # Print our report
    for item in items:
        print(
            "{}:{}{}".format(
                item[0], " " * ((longestName + minSpaces) - len(item[0])), item[1]
            )
        )



from .BuildConfiguration import BuildConfiguration
from .ContainerUtils import ContainerUtils
from .CredentialEndpoint import CredentialEndpoint
from .DarwinUtils import DarwinUtils
from .DockerUtils import DockerUtils
from .FilesystemUtils import FilesystemUtils
from .GlobalConfiguration import GlobalConfiguration
from .ImageBuilder import ImageBuilder
from .ImageCleaner import ImageCleaner
from .Logger import Logger
from .NetworkUtils import NetworkUtils
from .PrettyPrinting import PrettyPrinting
from .ResourceMonitor import ResourceMonitor
from .SubprocessUtils import SubprocessUtils
from .WindowsUtils import WindowsUtils



from .DockerUtils import DockerUtils
from packaging.version import Version
import platform, sys
from typing import Optional

if platform.system() == "Windows":
    import winreg


class WindowsUtils(object):
    # The oldest Windows build we support
    _minimumRequiredBuild = 17763

    # This lookup table is based on the list of valid tags from <https://hub.docker.com/r/microsoft/windowsservercore/>
    # and list of build-to-release mappings from the following pages:
    # - https://docs.microsoft.com/en-us/windows/release-health/release-information
    # - https://docs.microsoft.com/en-us/windows/release-health/windows11-release-information
    _knownTagsByBuildNumber = {
        17763: "ltsc2019",
        18362: "1903",
        18363: "1909",
        19041: "2004",
        19042: "20H2",
        19043: "21H1",
        20348: "ltsc2022",
        22000: "ltsc2022",
    }

    _knownTags = list(_knownTagsByBuildNumber.values())

    # The list of Windows Server and Windows 10 host OS releases that are blacklisted due to critical bugs
    # (See: <https://unrealcontainers.com/docs/concepts/windows-containers>)
    _blacklistedHosts = [18362, 18363]

    @staticmethod
    def _getVersionRegKey(subkey: str) -> str:
        """
        Retrieves the specified Windows version key from the registry

        @raises FileNotFoundError if registry key doesn't exist
        """
        key = winreg.OpenKey(
            winreg.HKEY_LOCAL_MACHINE, "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
        )
        value = winreg.QueryValueEx(key, subkey)
        winreg.CloseKey(key)
        return value[0]

    @staticmethod
    def requiredSizeLimit() -> float:
        """
        Returns the minimum required image size limit (in GB) for Windows containers
        """
        return 400.0

    @staticmethod
    def minimumRequiredBuild() -> int:
        """
        Returns the minimum required version of Windows 10 / Windows Server
        """
        return WindowsUtils._minimumRequiredBuild

    @staticmethod
    def systemString() -> str:
        """
        Generates a verbose human-readable version string for the Windows host system
        """
        return "{} (Build {}.{})".format(
            WindowsUtils._getVersionRegKey("ProductName"),
            WindowsUtils.getWindowsBuild(),
            WindowsUtils._getVersionRegKey("UBR"),
        )

    @staticmethod
    def getHostBaseTag() -> Optional[str]:
        """
        Retrieves the tag for the Windows Server Core base image matching the host Windows system
        """

        hostBuild = WindowsUtils.getWindowsBuild()

        return WindowsUtils._knownTagsByBuildNumber.get(hostBuild)

    @staticmethod
    def getWindowsBuild() -> int:
        """
        Returns build number for the Windows host system
        """
        return sys.getwindowsversion().build

    @staticmethod
    def isBlacklistedWindowsHost() -> bool:
        """
        Determines if host Windows version is one with bugs that make it unsuitable for use
        (defaults to checking the host OS release if one is not specified)
        """
        dockerVersion = Version(DockerUtils.version()["Version"])
        build = WindowsUtils.getWindowsBuild()
        return build in WindowsUtils._blacklistedHosts and dockerVersion < Version(
            "19.03.6"
        )

    @staticmethod
    def isWindowsServer() -> bool:
        """
        Determines if the Windows host system is Windows Server
        """
        # TODO: Replace this with something more reliable
        return "Windows Server" in WindowsUtils._getVersionRegKey("ProductName")

    @staticmethod
    def getDllSrcImage(basetag: str) -> str:
        """
        Returns Windows image that can be used as a source for DLLs missing from Windows Server Core base image
        """
        # TODO: we also need to use Windows Server image when user specifies custom tags, like '10.0.20348.169'
        image = {
            "ltsc2022": "mcr.microsoft.com/windows/server",
        }.get(basetag, "mcr.microsoft.com/windows")

        tag = {
            "ltsc2019": "1809",
        }.get(basetag, basetag)

        return f"{image}:{tag}"

    @staticmethod
    def getKnownBaseTags() -> [str]:
        """
        Returns the list of known tags for the Windows Server Core base image, in ascending chronological release order
        """
        return WindowsUtils._knownTags

    @staticmethod
    def isNewerBaseTag(older: str, newer: str) -> Optional[bool]:
        """
        Determines if the base tag `newer` is chronologically newer than the base tag `older`
        """
        try:
            return WindowsUtils._knownTags.index(newer) > WindowsUtils._knownTags.index(
                older
            )
        except ValueError:
            return None



#!/usr/bin/env python3
import os, re, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


def patchFile(filename, search, replace):
    contents = readFile(filename)
    patched = contents.replace(search, replace)
    writeFile(filename, patched)


# Apply our bugfixes to UnrealBuildTool (UBT)

# Ensure modules always have a binary output directory value set when exporting JSON
# (In some Engine versions, External modules do not have an output directory set, breaking modules that reference `$(BinaryOutputDir)`)
patchFile(
    os.path.join(sys.argv[1], "Configuration", "UEBuildTarget.cs"),
    "Module.ExportJson(Module.Binary?.OutputDir, GetExecutableDir(), Writer);",
    "Module.ExportJson((Module.Binary != null || Binaries.Count == 0) ? Module.Binary?.OutputDir : Binaries[0].OutputDir, GetExecutableDir(), Writer);",
)



from .DockerUtils import DockerUtils
import humanfriendly, os, subprocess, time


class ImageCleaner(object):
    def __init__(self, logger):
        self.logger = logger

    def clean(self, image, dryRun=False):
        """
        Removes the specified image
        """

        # Determine if we are running in "dry run" mode
        self.logger.action('Removing image "{}"...'.format(image))
        cleanCommand = ["docker", "rmi", image]
        if dryRun == True:
            print(cleanCommand)
        else:
            subprocess.call(cleanCommand)

    def cleanMultiple(self, images, dryRun=False):
        """
        Removes all of the images in the supplied list
        """
        for image in images:
            self.clean(image, dryRun)



#!/usr/bin/env python3
from os.path import dirname
from subprocess import run, PIPE
import json, re, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Determine whether a changelist override value was specified
changelistOverride = None
if len(sys.argv) > 2 and sys.argv[2] != "%CHANGELIST%":
    # If the override was "auto" then attempt to retrieve the CL number from the git commit message
    if sys.argv[2] == "auto":
        # Retrieve the commit message from git
        engineRoot = dirname(dirname(dirname(sys.argv[1])))
        commitMessage = run(
            ["git", "log", "-n", "1", "--format=%s%n%b"],
            cwd=engineRoot,
            stdout=PIPE,
            stderr=PIPE,
            universal_newlines=True,
        ).stdout.strip()

        # If the commit is a tagged engine release then it won't have a CL number, and using "auto" is user error
        if re.fullmatch("[0-9\\.]+ release", commitMessage) is not None:
            print(
                "Error: you are attempting to automatically retrieve the CL number for a tagged Unreal Engine release.\n"
                "For hotfix releases of the Unreal Engine, a CL override is not required and should not be specified.\n"
                "For supported .0 releases of the Unreal Engine, ue4-docker ships with known CL numbers, so an override should not be necessary.",
                file=sys.stderr,
            )
            sys.exit(1)

        # Attempt to extract the CL number from the commit message
        match = re.search("\\[CL ([0-9]+) by .+ in .+ branch\\]", commitMessage)
        if match is not None:
            changelistOverride = int(match.group(1))
        else:
            print(
                "Error: failed to find a CL number in the git commit message! This was the commit message:\n\n"
                + commitMessage,
                file=sys.stderr,
            )
            sys.exit(1)

    else:
        changelistOverride = int(sys.argv[2])

# Update the `Changelist` field to reflect the override if it was supplied, or else the `CompatibleChangelist` field in our version file
versionFile = sys.argv[1]
details = json.loads(readFile(versionFile))
details["Changelist"] = (
    changelistOverride
    if changelistOverride is not None
    else details["CompatibleChangelist"]
)
details["IsPromotedBuild"] = 1
patchedJson = json.dumps(details, indent=4)
writeFile(versionFile, patchedJson)
print("PATCHED BUILD.VERSION:\n{}".format(patchedJson), file=sys.stderr)



import os

from packaging.version import Version
from urllib.request import urlopen
import json

# The default namespace for our tagged container images
DEFAULT_TAG_NAMESPACE = "adamrehn"


class GlobalConfiguration(object):
    """
    Manages access to the global configuration settings for ue4-docker itself
    """

    @staticmethod
    def getLatestVersion():
        """
        Queries PyPI to determine the latest available release of ue4-docker
        """
        with urlopen("https://pypi.org/pypi/ue4-docker/json") as url:
            data = json.load(url)
            releases = [Version(release) for release in data["releases"]]
            return sorted(releases)[-1]

    @staticmethod
    def getTagNamespace():
        """
        Returns the currently-configured namespace for container image tags
        """
        return os.environ.get("UE4DOCKER_TAG_NAMESPACE", DEFAULT_TAG_NAMESPACE)

    @staticmethod
    def resolveTag(tag):
        """
        Resolves a Docker image tag with respect to our currently-configured namespace
        """

        # If the specified tag already includes a namespace, simply return it unmodified
        return (
            tag
            if "/" in tag
            else "{}/{}".format(GlobalConfiguration.getTagNamespace(), tag)
        )



from typing import Dict, Optional

from .DockerUtils import DockerUtils
from .FilesystemUtils import FilesystemUtils
from .GlobalConfiguration import GlobalConfiguration
import glob, humanfriendly, os, shutil, subprocess, tempfile, time
from os.path import basename, exists, join
from jinja2 import Environment


class ImageBuildParams(object):
    def __init__(
        self, dockerfile: str, context_dir: str, env: Optional[Dict[str, str]] = None
    ):
        self.dockerfile = dockerfile
        self.context_dir = context_dir
        self.env = env


class ImageBuilder(object):
    def __init__(
        self,
        tempDir: str,
        platform: str,
        logger,
        rebuild: bool = False,
        dryRun: bool = False,
        layoutDir: str = None,
        templateContext: Dict[str, str] = None,
        combine: bool = False,
    ):
        """
        Creates an ImageBuilder for the specified build parameters
        """
        self.tempDir = tempDir
        self.platform = platform
        self.logger = logger
        self.rebuild = rebuild
        self.dryRun = dryRun
        self.layoutDir = layoutDir
        self.templateContext = templateContext if templateContext is not None else {}
        self.combine = combine

    def get_built_image_context(self, name):
        """
        Resolve the full path to the build context for the specified image
        """
        return os.path.normpath(
            os.path.join(
                os.path.dirname(os.path.abspath(__file__)),
                "..",
                "dockerfiles",
                basename(name),
                self.platform,
            )
        )

    def build_builtin_image(
        self,
        name: str,
        tags: [str],
        args: [str],
        builtin_name: str = None,
        secrets: Dict[str, str] = None,
    ):
        context_dir = self.get_built_image_context(
            name if builtin_name is None else builtin_name
        )
        return self.build(
            name, tags, args, join(context_dir, "Dockerfile"), context_dir, secrets
        )

    def build(
        self,
        name: str,
        tags: [str],
        args: [str],
        dockerfile_template: str,
        context_dir: str,
        secrets: Dict[str, str] = None,
    ):
        """
        Builds the specified image if it doesn't exist or if we're forcing a rebuild
        """

        workdir = join(self.tempDir, basename(name), self.platform)
        os.makedirs(workdir, exist_ok=True)

        # Create a Jinja template environment and render the Dockerfile template
        environment = Environment(
            autoescape=False, trim_blocks=True, lstrip_blocks=True
        )
        dockerfile = join(workdir, "Dockerfile")

        templateInstance = environment.from_string(
            FilesystemUtils.readFile(dockerfile_template)
        )
        rendered = templateInstance.render(self.templateContext)

        # Compress excess whitespace introduced during Jinja rendering and save the contents back to disk
        # (Ensure that we still have a single trailing newline at the end of the Dockerfile)
        while "\n\n\n" in rendered:
            rendered = rendered.replace("\n\n\n", "\n\n")
        rendered = rendered.strip("\n") + "\n"
        FilesystemUtils.writeFile(dockerfile, rendered)

        # Inject our filesystem layer commit message after each RUN directive in the Dockerfile
        DockerUtils.injectPostRunMessage(
            dockerfile,
            self.platform,
            [
                "",
                "RUN directive complete. Docker will now commit the filesystem layer to disk.",
                "Note that for large filesystem layers this can take quite some time.",
                "Performing filesystem layer commit...",
                "",
            ],
        )

        # When building Linux images, explicitly specify the target CPU architecture
        archFlags = ["--platform", "linux/amd64"] if self.platform == "linux" else []

        # Create a temporary directory to hold any files needed for the build
        with tempfile.TemporaryDirectory() as tempDir:
            # Determine whether we are building using `docker buildx` with build secrets
            imageTags = self._formatTags(name, tags)

            if self.platform == "linux" and secrets is not None and len(secrets) > 0:
                # Create temporary files to store the contents of each of our secrets
                secretFlags = []
                for secret, contents in secrets.items():
                    secretFile = join(tempDir, secret)
                    FilesystemUtils.writeFile(secretFile, contents)
                    secretFlags.append("id={},src={}".format(secret, secretFile))

                # Generate the `docker buildx` command to use our build secrets
                command = DockerUtils.buildx(
                    imageTags, context_dir, archFlags + args, secretFlags
                )
            else:
                command = DockerUtils.build(imageTags, context_dir, archFlags + args)

            command += ["--file", dockerfile]

            env = os.environ.copy()
            if self.platform == "linux":
                env["DOCKER_BUILDKIT"] = "1"

            # Build the image if it doesn't already exist
            self._processImage(
                imageTags[0],
                name,
                command,
                "build",
                "built",
                ImageBuildParams(dockerfile, context_dir, env),
            )

    def pull(self, image: str) -> None:
        """
        Pulls the specified image if it doesn't exist or if we're forcing a pull of a newer version
        """
        self._processImage(image, None, DockerUtils.pull(image), "pull", "pulled")

    def willBuild(self, name: str, tags: [str]) -> bool:
        """
        Determines if we will build the specified image, based on our build settings
        """
        imageTags = self._formatTags(name, tags)
        return self._willProcess(imageTags[0])

    def _formatTags(self, name: str, tags: [str]):
        """
        Generates the list of fully-qualified tags that we will use when building an image
        """
        return [
            "{}:{}".format(GlobalConfiguration.resolveTag(name), tag) for tag in tags
        ]

    def _willProcess(self, image: [str]) -> bool:
        """
        Determines if we will build or pull the specified image, based on our build settings
        """
        return self.rebuild or not DockerUtils.exists(image)

    def _processImage(
        self,
        image: str,
        name: Optional[str],
        command: [str],
        actionPresentTense: str,
        actionPastTense: str,
        build_params: Optional[ImageBuildParams] = None,
    ) -> None:
        """
        Processes the specified image by running the supplied command if it doesn't exist (use rebuild=True to force processing)
        """

        # Determine if we are processing the image
        if not self._willProcess(image):
            self.logger.info(
                'Image "{}" exists and rebuild not requested, skipping {}.'.format(
                    image, actionPresentTense
                )
            )
            return

        # Determine if we are running in "dry run" mode
        self.logger.action(
            '{}ing image "{}"...'.format(actionPresentTense.capitalize(), image)
        )
        if self.dryRun:
            print(command)
            self.logger.action(
                'Completed dry run for image "{}".'.format(image), newline=False
            )
            return

        # Determine if we're just copying the Dockerfile to an output directory
        if self.layoutDir is not None:
            # Determine whether we're performing a simple copy or combining generated Dockerfiles
            if self.combine:
                # Ensure the destination directory exists
                dest = join(self.layoutDir, "combined")
                self.logger.action(
                    'Merging "{}" into "{}"...'.format(build_params.context_dir, dest),
                    newline=False,
                )
                os.makedirs(dest, exist_ok=True)

                # Merge the source Dockerfile with any existing Dockerfile contents in the destination directory
                # (Insert a single newline between merged file contents and ensure we have a single trailing newline)
                sourceDockerfile = build_params.dockerfile
                destDockerfile = join(dest, "Dockerfile")
                dockerfileContents = (
                    FilesystemUtils.readFile(destDockerfile)
                    if exists(destDockerfile)
                    else ""
                )
                dockerfileContents = (
                    dockerfileContents
                    + "\n"
                    + FilesystemUtils.readFile(sourceDockerfile)
                )
                dockerfileContents = dockerfileContents.strip("\n") + "\n"
                FilesystemUtils.writeFile(destDockerfile, dockerfileContents)

                # Copy any supplemental files from the source directory to the destination directory
                # (Exclude any extraneous files which are not referenced in the Dockerfile contents)
                for file in glob.glob(join(build_params.context_dir, "*.*")):
                    if basename(file) in dockerfileContents:
                        shutil.copy(file, join(dest, basename(file)))

                # Report our success
                self.logger.action(
                    'Merged Dockerfile for image "{}".'.format(image), newline=False
                )

            else:
                # Copy the source directory to the destination
                dest = join(self.layoutDir, basename(name))
                self.logger.action(
                    'Copying "{}" to "{}"...'.format(build_params.context_dir, dest),
                    newline=False,
                )
                shutil.copytree(build_params.context_dir, dest)
                shutil.copy(build_params.dockerfile, dest)
                self.logger.action(
                    'Copied Dockerfile for image "{}".'.format(image), newline=False
                )

            return

        # Attempt to process the image using the supplied command
        startTime = time.time()
        exitCode = subprocess.call(
            command, env=build_params.env if build_params else None
        )
        endTime = time.time()

        # Determine if processing succeeded
        if exitCode == 0:
            self.logger.action(
                '{} image "{}" in {}'.format(
                    actionPastTense.capitalize(),
                    image,
                    humanfriendly.format_timespan(endTime - startTime),
                ),
                newline=False,
            )
        else:
            raise RuntimeError(
                'failed to {} image "{}".'.format(actionPresentTense, image)
            )



from .build import build
from .clean import clean
from .export import export
from .info import info
from .main import main
from .setup_cmd import setup
from .test import test
from .version_cmd import version
from .version import __version__



class PrettyPrinting(object):
    @staticmethod
    def printColumns(pairs, indent=2, minSpaces=6):
        """
        Prints a list of paired values in two nicely aligned columns
        """

        # Determine the length of the longest item in the left-hand column
        longestName = max([len(pair[0]) for pair in pairs])

        # Print the two columns
        for pair in pairs:
            whitespace = " " * ((longestName + minSpaces) - len(pair[0]))
            print("{}{}{}{}".format(" " * indent, pair[0], whitespace, pair[1]))



#!/usr/bin/env python3
import os, platform, subprocess, tempfile, ue4cli


# Runs a command, raising an error if it returns a nonzero exit code
def run(command, **kwargs):
    print("[RUN COMMAND] {} {}".format(command, kwargs), flush=True)
    return subprocess.run(command, check=True, **kwargs)


# Retrieve the short version string for the Engine
manager = ue4cli.UnrealManagerFactory.create()
version = manager.getEngineVersion("short")

# Create an auto-deleting temporary directory to work in
with tempfile.TemporaryDirectory() as tempDir:
    # Clone a simple C++ project and verify that we can build and package it
    repo = "https://gitlab.com/ue4-test-projects/{}/BasicCxx.git".format(version)
    projectDir = os.path.join(tempDir, "BasicCxx")
    run(["git", "clone", "--depth=1", repo, projectDir])
    run(["ue4", "package", "Shipping"], cwd=projectDir)

    # Forcibly delete the .git subdirectory under Windows to avoid permissions errors when deleting the temp directory
    if platform.system() == "Windows":
        run(["del", "/f", "/s", "/q", os.path.join(projectDir, ".git")], shell=True)



#!/usr/bin/env python3
import os, platform, subprocess, tempfile, ue4cli


# Reads data from a file
def read(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


# Runs a command, raising an error if it returns a nonzero exit code
def run(command, **kwargs):
    print("[RUN COMMAND] {} {}".format(command, kwargs), flush=True)
    return subprocess.run(command, check=True, **kwargs)


# Writes data to a file
def write(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Retrieve the short version string for the Engine
manager = ue4cli.UnrealManagerFactory.create()
version = manager.getEngineVersion("short")

# Create an auto-deleting temporary directory to work in
with tempfile.TemporaryDirectory() as tempDir:
    # Clone a simple C++ project
    repo = "https://gitlab.com/ue4-test-projects/{}/BasicCxx.git".format(version)
    projectDir = os.path.join(tempDir, "BasicCxx")
    run(["git", "clone", "--depth=1", repo, projectDir])

    # Generate a code module to wrap our external dependencies
    sourceDir = os.path.join(projectDir, "Source")
    run(["ue4", "conan", "boilerplate", "WrapperModule"], cwd=sourceDir)

    # Add the wrapper module as a dependency of the project's main source code module
    rulesFile = os.path.join(sourceDir, "BasicCxx", "BasicCxx.Build.cs")
    rules = read(rulesFile)
    rules = rules.replace(
        "PublicDependencyModuleNames.AddRange(new string[] {",
        'PublicDependencyModuleNames.AddRange(new string[] { "WrapperModule", ',
    )
    write(rulesFile, rules)

    # Add some dependencies to the module's conanfile.py
    moduleDir = os.path.join(sourceDir, "WrapperModule")
    conanfile = os.path.join(moduleDir, "conanfile.py")
    deps = read(conanfile)
    deps = deps.replace("pass", 'self._requireUnreal("zlib/ue4@adamrehn/{}")')
    write(conanfile, deps)

    # Verify that we can build the project with dynamically located dependencies
    run(["ue4", "build"], cwd=projectDir)
    run(["ue4", "clean"], cwd=projectDir)

    # Verify that we can build the project with precomputed dependency data
    run(["ue4", "conan", "precompute", "host"], cwd=moduleDir)
    run(["ue4", "build"], cwd=projectDir)

    # Forcibly delete the .git subdirectory under Windows to avoid permissions errors when deleting the temp directory
    if platform.system() == "Windows":
        run(["del", "/f", "/s", "/q", os.path.join(projectDir, ".git")], shell=True)



#!/usr/bin/env python3
import os, re, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Ensure the `PlatformType` field is set correctly for Client and Server targets in BaseEngine.ini
iniFile = sys.argv[1]
config = readFile(iniFile)
config = re.sub(
    'PlatformType="Game", RequiredFile="(.+UE4(Client|Server).*\\.target)"',
    'PlatformType="\\2", RequiredFile="\\1"',
    config,
)
writeFile(iniFile, config)



#!/usr/bin/env python3
import json, os, subprocess, sys
from os.path import join


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Determine if we are building UE 4.25.4
engineRoot = sys.argv[1]
verboseOutput = len(sys.argv) > 2 and sys.argv[2] == "1"
versionDetails = json.loads(
    readFile(join(engineRoot, "Engine", "Build", "Build.version"))
)
if (
    versionDetails["MajorVersion"] == 4
    and versionDetails["MinorVersion"] == 25
    and versionDetails["PatchVersion"] == 4
):
    # If `Commit.gitdeps.xml` is missing the changes from CL 14469950 then inject them
    # (See: <https://github.com/EpicGames/UnrealEngine/commit/84e4ea3241c294c04fdf7d8fb63f99a3109c8edd>)
    gitdepsFile = join(engineRoot, "Engine", "Build", "Commit.gitdeps.xml")
    gitdepsXml = readFile(gitdepsFile)
    if '<File Name="cpp.hint"' not in gitdepsXml:
        gitdepsXml = gitdepsXml.replace(
            '<File Name=".tgitconfig" Hash="d3d7bbcf9b2fc8b6e4f2965354a5633c4f175589" />',
            '<File Name=".tgitconfig" Hash="d3d7bbcf9b2fc8b6e4f2965354a5633c4f175589" />\n    '
            + '<File Name="cpp.hint" Hash="7d1daec3c6218ce9f49f9be0280091b98d7168d7" />',
        )
        gitdepsXml = gitdepsXml.replace(
            '<Blob Hash="7d1492e46d159b6979f70a415727a2be7e569e21" Size="342112" PackHash="feb61b7040721b885ad85174cfc802419600bda1" PackOffset="1545471" />',
            '<Blob Hash="7d1492e46d159b6979f70a415727a2be7e569e21" Size="342112" PackHash="feb61b7040721b885ad85174cfc802419600bda1" PackOffset="1545471" />\n    '
            + '<Blob Hash="7d1daec3c6218ce9f49f9be0280091b98d7168d7" Size="456" PackHash="33e382aea05629bd179a60cf1520f77c025ac0b3" PackOffset="8" />',
        )
        gitdepsXml = gitdepsXml.replace(
            '<Pack Hash="33d0a2949662b327b35a881192e85107ecafc8ac" Size="2097152" CompressedSize="655885" RemotePath="2369826-2acd3c361c9d4a858bd63938a2ab980e" />',
            '<Pack Hash="33d0a2949662b327b35a881192e85107ecafc8ac" Size="2097152" CompressedSize="655885" RemotePath="2369826-2acd3c361c9d4a858bd63938a2ab980e" />\n    '
            + '<Pack Hash="33e382aea05629bd179a60cf1520f77c025ac0b3" Size="464" CompressedSize="235" RemotePath="UnrealEngine-14572338-da4318c3ab684bc48601f32f0b1b6fe3" />',
        )

        writeFile(gitdepsFile, gitdepsXml)

        if verboseOutput == True:
            print("PATCHED {}:\n\n{}".format(gitdepsFile, gitdepsXml), file=sys.stderr)
        else:
            print("PATCHED {}".format(gitdepsFile), file=sys.stderr)



import ntpath
import os
import posixpath
import sys

import docker
from docker.errors import ImageNotFound

from .infrastructure import (
    ContainerUtils,
    GlobalConfiguration,
    Logger,
)


def test():
    # Create our logger to generate coloured output on stderr
    logger = Logger(prefix="[{} test] ".format(sys.argv[0]))

    # Create our Docker API client
    client = docker.from_env()

    # Check that an image tag has been specified
    if len(sys.argv) > 1 and sys.argv[1].strip("-") not in ["h", "help"]:
        # Verify that the specified container image exists
        tag = sys.argv[1]
        image_name = GlobalConfiguration.resolveTag(
            "ue4-full:{}".format(tag) if ":" not in tag else tag
        )

        try:
            image = client.images.get(image_name)
        except ImageNotFound:
            logger.error(
                'Error: the specified container image "{}" does not exist.'.format(
                    image_name
                )
            )
            sys.exit(1)

        # Use process isolation mode when testing Windows containers, since running Hyper-V containers don't currently support manipulating the filesystem
        platform = image.attrs["Os"]
        isolation = "process" if platform == "windows" else None

        # Start a container to run our tests in, automatically stopping and removing the container when we finish
        logger.action(
            'Starting a container using the "{}" image...'.format(image_name), False
        )
        container = ContainerUtils.start_for_exec(
            client, image_name, platform, isolation=isolation
        )
        with ContainerUtils.automatically_stop(container):
            # Create the workspace directory in the container
            workspaceDir = (
                "C:\\workspace" if platform == "windows" else "/tmp/workspace"
            )
            shell_prefix = (
                ["cmd", "/S", "/C"] if platform == "windows" else ["bash", "-c"]
            )

            ContainerUtils.exec(
                container,
                shell_prefix + ["mkdir " + workspaceDir],
            )

            # Copy our test scripts into the container
            testDir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "tests")
            ContainerUtils.copy_from_host(container, testDir, workspaceDir)

            # Create a harness to invoke individual tests
            containerPath = ntpath if platform == "windows" else posixpath
            pythonCommand = "python" if platform == "windows" else "python3"

            def runTest(script):
                logger.action('Running test "{}"...'.format(script), False)
                try:
                    ContainerUtils.exec(
                        container,
                        [pythonCommand, containerPath.join(workspaceDir, script)],
                        workdir=workspaceDir,
                    )
                    logger.action('Passed test "{}"'.format(script), False)
                except RuntimeError as e:
                    logger.error('Error: test "{}" failed!'.format(script))
                    raise e from None

            # Run each of our tests in turn
            runTest("build-and-package.py")
            runTest("consume-external-deps.py")

            # If we've reached this point then all of the tests passed
            logger.action("All tests passed.", False)

    else:
        # Print usage syntax
        print("Usage: {} test TAG".format(sys.argv[0]))
        print("Runs tests to verify the correctness of built container images\n")
        print("TAG should specify the tag of the ue4-full image to test.")



from .infrastructure import DockerUtils, GlobalConfiguration, PrettyPrinting
from .exports import *
import sys


def _notNone(items):
    return len([i for i in items if i is not None]) == len(items)


def _extractArg(args, index):
    return args[index] if len(args) > index else None


def _isHelpFlag(arg):
    return (arg.strip("-") in ["h", "help"]) == True


def _stripHelpFlags(args):
    return list([a for a in args if _isHelpFlag(a) == False])


def export():
    # The components that can be exported
    COMPONENTS = {
        "installed": {
            "function": exportInstalledBuild,
            "description": "Exports an Installed Build of the Engine",
            "image": GlobalConfiguration.resolveTag("ue4-full"),
            "help": "Copies the Installed Build from a container to the host system.\nOnly supported under Linux for UE 4.21.0 and newer.",
        },
        "packages": {
            "function": exportPackages,
            "description": "Exports conan-ue4cli wrapper packages",
            "image": GlobalConfiguration.resolveTag("ue4-full"),
            "help": "Runs a temporary conan server inside a container and uses it to export the\ngenerated conan-ue4cli wrapper packages.\n\n"
            + 'Currently the only supported destination value is "cache", which exports\nthe packages to the Conan local cache on the host system.',
        },
    }

    # Parse the supplied command-line arguments
    stripped = _stripHelpFlags(sys.argv)
    args = {
        "help": len(stripped) < len(sys.argv),
        "component": _extractArg(stripped, 1),
        "tag": _extractArg(stripped, 2),
        "destination": _extractArg(stripped, 3),
    }

    # If a component name has been specified, verify that it is valid
    if args["component"] is not None and args["component"] not in COMPONENTS:
        print(
            'Error: unrecognised component "{}".'.format(args["component"]),
            file=sys.stderr,
        )
        sys.exit(1)

    # Determine if we are performing an export
    if args["help"] == False and _notNone(
        [args["component"], args["tag"], args["destination"]]
    ):
        # Determine if the user specified an image and a tag or just a tag
        tag = args["tag"]
        details = COMPONENTS[args["component"]]
        requiredImage = "{}:{}".format(details["image"], tag) if ":" not in tag else tag

        # Verify that the required container image exists
        if DockerUtils.exists(requiredImage) == False:
            print(
                'Error: the specified container image "{}" does not exist.'.format(
                    requiredImage
                ),
                file=sys.stderr,
            )
            sys.exit(1)

        # Attempt to perform the export
        details["function"](requiredImage, args["destination"], stripped[4:])
        print("Export complete.")

    # Determine if we are displaying the help for a specific component
    elif args["help"] == True and args["component"] is not None:
        # Display the help for the component
        component = sys.argv[1]
        details = COMPONENTS[component]
        print("{} export {}".format(sys.argv[0], component))
        print(details["description"] + "\n")
        print("Exports from image: {}:TAG\n".format(details["image"]))
        print(details["help"])

    else:
        # Print usage syntax
        print("Usage: {} export COMPONENT TAG DESTINATION\n".format(sys.argv[0]))
        print("Exports components from built container images to the host system\n")
        print("Components:")
        PrettyPrinting.printColumns(
            [
                (component, COMPONENTS[component]["description"])
                for component in COMPONENTS
            ]
        )
        print(
            "\nRun `{} export COMPONENT --help` for more information on a component.".format(
                sys.argv[0]
            )
        )



#!/usr/bin/env python3
import os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


filtersXml = sys.argv[1]
code = readFile(filtersXml)

# Add missing SetupDotnet.sh in Unreal Engine 5.0.0-early-access-1
# See https://github.com/adamrehn/ue4-docker/issues/171#issuecomment-853918412
# and https://github.com/EpicGames/UnrealEngine/commit/a18824057e6cd490750a10b59af29ca10b3d67d9
dotnet = "Engine/Binaries/ThirdParty/DotNet/Linux/..."
setup_dotnet = "Engine/Build/BatchFiles/Linux/SetupDotnet.sh"
if dotnet in code and setup_dotnet not in code:
    code = code.replace(dotnet, f"{dotnet}\n{setup_dotnet}")

writeFile(filtersXml, code)



from .infrastructure import Logger, PrettyPrinting
from .diagnostics import *
import sys


def diagnostics():
    # The diagnostics that can be run
    DIAGNOSTICS = {
        "all": allDiagnostics(),
        "8gig": diagnostic8Gig(),
        "20gig": diagnostic20Gig(),
        "network": diagnosticNetwork(),
    }

    # Create our logger to generate coloured output on stderr
    logger = Logger(prefix="[{} diagnostics] ".format(sys.argv[0]))

    # Parse the supplied command-line arguments
    stripped = list([arg for arg in sys.argv if arg.strip("-") not in ["h", "help"]])
    args = {
        "help": len(sys.argv) > len(stripped),
        "diagnostic": stripped[1] if len(stripped) > 1 else None,
    }

    # If a diagnostic name has been specified, verify that it is valid
    if args["diagnostic"] is not None and args["diagnostic"] not in DIAGNOSTICS:
        logger.error(
            'Error: unrecognised diagnostic "{}".'.format(args["diagnostic"]), False
        )
        sys.exit(1)

    # Determine if we are running a diagnostic
    if args["help"] == False and args["diagnostic"] is not None:
        # Run the diagnostic
        diagnostic = DIAGNOSTICS[args["diagnostic"]]
        logger.action('Running diagnostic: "{}"'.format(diagnostic.getName()), False)
        passed = diagnostic.run(logger, stripped[2:])

        # Print the result
        if passed == True:
            logger.action("Diagnostic result: passed", False)
        else:
            logger.error("Diagnostic result: failed", False)

    # Determine if we are displaying the help for a specific diagnostic
    elif args["help"] == True and args["diagnostic"] is not None:
        # Display the help for the diagnostic
        diagnostic = DIAGNOSTICS[args["diagnostic"]]
        print("{} diagnostics {}".format(sys.argv[0], args["diagnostic"]))
        print(diagnostic.getName() + "\n")
        print(diagnostic.getDescription())

    else:
        # Print usage syntax
        print("Usage: {} diagnostics DIAGNOSTIC\n".format(sys.argv[0]))
        print("Runs diagnostics to detect issues with the host system configuration\n")
        print("Available diagnostics:")
        PrettyPrinting.printColumns(
            [
                (diagnostic, DIAGNOSTICS[diagnostic].getName())
                for diagnostic in DIAGNOSTICS
            ]
        )
        print(
            "\nRun `{} diagnostics DIAGNOSTIC --help` for more information on a diagnostic.".format(
                sys.argv[0]
            )
        )



import socket


class NetworkUtils(object):
    @staticmethod
    def hostIP():
        """
        Determines the IP address of the host
        """
        # Code from <https://stackoverflow.com/a/28950776>
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(("10.255.255.255", 1))
            IP = s.getsockname()[0]
        except:
            IP = "127.0.0.1"
        finally:
            s.close()
        return IP



from ..infrastructure import DockerUtils, WindowsUtils
from .base import DiagnosticBase

import argparse, os, platform
from os.path import abspath, dirname, join


class diagnostic20Gig(DiagnosticBase):
    # The tag we use for built images
    IMAGE_TAG = "adamrehn/ue4-docker/diagnostics:20gig"

    def __init__(self):
        # Setup our argument parser so we can use its help message output in our description text
        self._parser = argparse.ArgumentParser(prog="ue4-docker diagnostics 20gig")
        self._parser.add_argument(
            "--isolation",
            default=None,
            choices=["hyperv", "process"],
            help="Override the default isolation mode when testing Windows containers",
        )
        self._parser.add_argument(
            "-basetag",
            default=None,
            choices=WindowsUtils.getKnownBaseTags(),
            help="Override the default base image tag when testing Windows containers",
        )

    def getName(self):
        """
        Returns the human-readable name of the diagnostic
        """
        return "Check for Docker 20GiB COPY bug"

    def getDescription(self):
        """
        Returns a description of what the diagnostic does
        """
        return "\n".join(
            [
                "This diagnostic determines if the Docker daemon suffers from 20GiB COPY bug",
                "",
                "See https://github.com/adamrehn/ue4-docker/issues/99#issuecomment-1079702817 for details and workarounds",
                "",
                self._parser.format_help(),
            ]
        )

    def getPrefix(self):
        """
        Returns the short name of the diagnostic for use in log output
        """
        return "20gig"

    def run(self, logger, args=[]):
        """
        Runs the diagnostic
        """

        # Parse our supplied arguments
        args = self._parser.parse_args(args)

        # Determine which platform we are running on
        containerPlatform = platform.system().lower()

        if containerPlatform != "windows":
            logger.action(
                "[20gig] Diagnostic skipped. Current platform is not affected by the bug this diagnostic checks\n"
            )
            return True

        buildArgs = self._generateWindowsBuildArgs(logger, args.basetag, args.isolation)

        # Attempt to build the Dockerfile
        logger.action(
            "[20gig] Attempting to COPY more than 20GiB between layers...", False
        )
        built = self._buildDockerfile(
            logger, containerPlatform, diagnostic20Gig.IMAGE_TAG, buildArgs
        )

        # Inform the user of the outcome of the diagnostic
        if built == True:
            logger.action(
                "[20gig] Diagnostic succeeded! The Docker daemon can COPY more than 20GiB between layers.\n"
            )
        else:
            logger.error(
                "[20gig] Diagnostic failed! The Docker daemon cannot COPY more than 20GiB between layers.\n",
                True,
            )

        return built



#!/usr/bin/env python3
from os.path import basename, dirname, exists, join, relpath
import glob, shutil, sys

# Determine the root directory for the source build and the Installed Build
sourceRoot = sys.argv[1]
installedRoot = join(sourceRoot, "LocalBuilds", "Engine", "Linux")

# Locate the bundled toolchain and copy it to the Installed Build
sdkGlob = join(
    sourceRoot,
    "Engine",
    "Extras",
    "ThirdPartyNotUE",
    "SDKs",
    "HostLinux",
    "Linux_x64",
    "*",
    "x86_64-unknown-linux-gnu",
)
for bundled in glob.glob(sdkGlob):
    # Extract the root path for the toolchain
    toolchain = dirname(bundled)

    # Print progress output
    print(
        'Copying bundled toolchain "{}" to Installed Build...'.format(
            basename(toolchain)
        ),
        file=sys.stderr,
    )
    sys.stderr.flush()

    # Perform the copy
    dest = join(installedRoot, relpath(toolchain, sourceRoot))
    if exists(dest) == True:
        print(
            "Destination toolchain already exists: {}".format(dest),
            file=sys.stderr,
            flush=True,
        )
    else:
        shutil.copytree(toolchain, dest)



from .infrastructure import (
    DarwinUtils,
    DockerUtils,
    Logger,
    PrettyPrinting,
    WindowsUtils,
)
from .build import build
from .clean import clean
from .diagnostics_cmd import diagnostics
from .export import export
from .info import info
from .setup_cmd import setup
from .test import test
from .version_cmd import version
import logging, os, platform, sys


def _exitWithError(err):
    Logger().error(err)
    sys.exit(1)


def main():
    # Configure verbose logging if the user requested it
    # (NOTE: in a future version of ue4-docker the `Logger` class will be properly integrated with standard logging)
    if "-v" in sys.argv or "--verbose" in sys.argv:
        # Enable verbose logging
        logging.getLogger().setLevel(logging.DEBUG)

        # Filter out the verbose flag to avoid breaking commands that don't support it
        if not (len(sys.argv) > 1 and sys.argv[1] in ["build"]):
            sys.argv = list([arg for arg in sys.argv if arg not in ["-v", "--verbose"]])

    # Verify that Docker is installed
    installed, error = DockerUtils.installed()
    if installed == False:
        _exitWithError(
            "Error: could not detect Docker daemon version. Please ensure Docker is installed.\n\nError details: {}".format(
                error
            )
        )

    # Under Windows, verify that the host is a supported version
    if platform.system() == "Windows":
        host_build = WindowsUtils.getWindowsBuild()
        min_build = WindowsUtils.minimumRequiredBuild()
        if host_build < min_build:
            _exitWithError(
                "Error: the detected build of Windows ({}) is not supported. {} or newer is required.".format(
                    host_build, min_build
                )
            )

    # Under macOS, verify that the host is a supported version
    if platform.system() == "Darwin" and DarwinUtils.isSupportedMacOsVersion() == False:
        _exitWithError(
            "Error: the detected version of macOS ({}) is not supported. macOS {} or newer is required.".format(
                DarwinUtils.getMacOsVersion(), DarwinUtils.minimumRequiredVersion()
            )
        )

    # Our supported commands
    COMMANDS = {
        "build": {
            "function": build,
            "description": "Builds container images for UE4",
        },
        "clean": {"function": clean, "description": "Cleans built container images"},
        "diagnostics": {
            "function": diagnostics,
            "description": "Runs diagnostics to detect issues with the host system configuration",
        },
        "export": {
            "function": export,
            "description": "Exports components from built container images to the host system",
        },
        "info": {
            "function": info,
            "description": "Displays information about the host system and Docker daemon",
        },
        "setup": {
            "function": setup,
            "description": "Automatically configures the host system where possible",
        },
        "test": {
            "function": test,
            "description": "Runs tests to verify the correctness of built container images",
        },
        "version": {
            "function": version,
            "description": "Prints the ue4-docker version number",
        },
    }

    # Truncate argv[0] to just the command name without the full path
    sys.argv[0] = os.path.basename(sys.argv[0])

    # Determine if a command has been specified
    if len(sys.argv) > 1:
        # Verify that the specified command is valid
        command = sys.argv[1]
        if command not in COMMANDS:
            print('Error: unrecognised command "{}".'.format(command), file=sys.stderr)
            sys.exit(1)

        # Invoke the command
        sys.argv = [sys.argv[0]] + sys.argv[2:]
        COMMANDS[command]["function"]()

    else:
        # Print usage syntax
        print("Usage: {} COMMAND [OPTIONS]\n".format(sys.argv[0]))
        print("Windows and Linux containers for Unreal Engine 4\n")
        print("Commands:")
        PrettyPrinting.printColumns(
            [(command, COMMANDS[command]["description"]) for command in COMMANDS]
        )
        print(
            "\nRun `{} COMMAND --help` for more information on a command.".format(
                sys.argv[0]
            )
        )



#!/usr/bin/env python3
import os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Remove the dependency on Linux cross-compilation debug tools introduced in UE4.20.0
filtersXml = sys.argv[1]
code = readFile(filtersXml)
code = code.replace("Engine/Binaries/Linux/dump_syms.exe", "")
code = code.replace("Engine/Binaries/Linux/BreakpadSymbolEncoder.exe", "")
writeFile(filtersXml, code)



#!/usr/bin/env python3
import glob, os, shutil, sys

# Retrieve the Linux SDK root directory and system ld location from our command-line arguments
sdkRoot = sys.argv[1]
systemLd = sys.argv[2]

# Locate the bundled version(s) of ld and replace them with symlinks to the system ld
for bundled in glob.glob(
    os.path.join(
        sdkRoot, "*", "x86_64-unknown-linux-gnu", "bin", "x86_64-unknown-linux-gnu-ld"
    )
):
    os.unlink(bundled)
    os.symlink(systemLd, bundled)
    print("{} => {}".format(bundled, systemLd), file=sys.stderr)



from ..infrastructure import DockerUtils, FilesystemUtils, Logger, SubprocessUtils
import docker, os, subprocess, sys, tempfile

# The name we use for our temporary Conan remote
REMOTE_NAME = "_ue4docker_export_temp"

# Our conan_server config file data
CONAN_SERVER_CONFIG = """
[server]
jwt_secret: jwt_secret
jwt_expire_minutes: 120
ssl_enabled: False
port: 9300
public_port: 9300
host_name: {}
authorize_timeout: 1800
disk_storage_path: {}
disk_authorize_timeout: 1800
updown_secret: updown_secret

[write_permissions]
*/*@*/*: *

[read_permissions]
*/*@*/*: *

[users]
user: password
"""


def exportPackages(image, destination, extraArgs):
    # Create our logger to generate coloured output on stderr
    logger = Logger()

    # Verify that the destination is "cache"
    if destination.lower() != "cache":
        logger.error('Error: the only supported package export destination is "cache".')
        sys.exit(1)

    # Verify that Conan is installed on the host
    try:
        SubprocessUtils.run(["conan", "--version"])
    except:
        logger.error(
            "Error: Conan must be installed on the host system to export packages."
        )
        sys.exit(1)

    # Determine if the container image is a Windows image or a Linux image
    imageOS = DockerUtils.listImages(image)[0].attrs["Os"]

    # Use the appropriate commands and paths for the container platform
    cmdsAndPaths = {
        "linux": {
            "rootCommand": ["bash", "-c", "sleep infinity"],
            "mkdirCommand": ["mkdir"],
            "copyCommand": ["cp", "-f"],
            "dataDir": "/home/ue4/.conan_server/data",
            "configDir": "/home/ue4/.conan_server/",
            "bindMount": "/hostdir/",
        },
        "windows": {
            "rootCommand": ["timeout", "/t", "99999", "/nobreak"],
            "mkdirCommand": ["cmd", "/S", "/C", "mkdir"],
            "copyCommand": ["xcopy", "/y"],
            "dataDir": "C:\\Users\\ContainerAdministrator\\.conan_server\\data",
            "configDir": "C:\\Users\\ContainerAdministrator\\.conan_server\\",
            "bindMount": "C:\\hostdir\\",
        },
    }[imageOS]

    # Create an auto-deleting temporary directory to hold our server config file
    with tempfile.TemporaryDirectory() as tempDir:
        # Progress output
        print("Starting conan_server in a container...")

        # Start a container from which we will export packages, bind-mounting our temp directory
        container = DockerUtils.start(
            image,
            cmdsAndPaths["rootCommand"],
            ports={"9300/tcp": 9300},
            mounts=[docker.types.Mount(cmdsAndPaths["bindMount"], tempDir, "bind")],
            stdin_open=imageOS == "windows",
            tty=imageOS == "windows",
            remove=True,
        )

        # Reload the container attributes from the Docker daemon to ensure the networking fields are populated
        container.reload()

        # Under Linux we can simply access the container from the host over the loopback address, but this doesn't work under Windows
        # (See <https://blog.sixeyed.com/published-ports-on-windows-containers-dont-do-loopback/>)
        externalAddress = (
            "127.0.0.1"
            if imageOS == "linux"
            else container.attrs["NetworkSettings"]["Networks"]["nat"]["IPAddress"]
        )

        # Generate our server config file in the temp directory
        FilesystemUtils.writeFile(
            os.path.join(tempDir, "server.conf"),
            CONAN_SERVER_CONFIG.format(externalAddress, cmdsAndPaths["dataDir"]),
        )

        # Keep track of the `conan_server` log output so we can display it in case of an error
        serverOutput = None

        try:
            # Copy the server config file to the expected location inside the container
            DockerUtils.execMultiple(
                container,
                [
                    cmdsAndPaths["mkdirCommand"] + [cmdsAndPaths["configDir"]],
                    cmdsAndPaths["copyCommand"]
                    + [
                        cmdsAndPaths["bindMount"] + "server.conf",
                        cmdsAndPaths["configDir"],
                    ],
                ],
            )

            # Start `conan_server`
            serverOutput = DockerUtils.exec(container, ["conan_server"], stream=True)

            # Progress output
            print("Uploading packages to the server...")

            # Upload all of the packages in the container's local cache to the server
            DockerUtils.execMultiple(
                container,
                [
                    ["conan", "remote", "add", "localhost", "http://127.0.0.1:9300"],
                    ["conan", "user", "user", "-r", "localhost", "-p", "password"],
                    ["conan", "upload", "*/4.*", "--all", "--confirm", "-r=localhost"],
                ],
            )

            # Configure the server as a temporary remote on the host system
            SubprocessUtils.run(
                [
                    "conan",
                    "remote",
                    "add",
                    REMOTE_NAME,
                    "http://{}:9300".format(externalAddress),
                ]
            )
            SubprocessUtils.run(
                ["conan", "user", "user", "-r", REMOTE_NAME, "-p", "password"]
            )

            # Retrieve the list of packages that were uploaded to the server
            packages = SubprocessUtils.extractLines(
                SubprocessUtils.capture(
                    ["conan", "search", "-r", REMOTE_NAME, "*"]
                ).stdout
            )
            packages = [
                package for package in packages if "/" in package and "@" in package
            ]

            # Download each package in turn
            for package in packages:
                print(
                    "Downloading package {} to host system local cache...".format(
                        package
                    )
                )
                SubprocessUtils.run(["conan", "download", "-r", REMOTE_NAME, package])

            # Once we reach this point, everything has worked and we don't need to output any logs
            serverOutput = None

        finally:
            # Stop the container, irrespective of whether or not the export succeeded
            print("Stopping conan_server...")
            container.stop()

            # If something went wrong then output the logs from `conan_server` to assist in diagnosing the failure
            if serverOutput is not None:
                print("Log output from conan_server:")
                for chunk in serverOutput:
                    logger.error(chunk.decode("utf-8"))

            # Remove the temporary remote if it was created successfully
            SubprocessUtils.run(["conan", "remote", "remove", REMOTE_NAME], check=False)



from termcolor import colored
import colorama, sys


class Logger(object):
    def __init__(self, prefix=""):
        """
        Creates a logger that will print coloured output to stderr
        """
        colorama.init()
        self.prefix = prefix

    def action(self, output, newline=True):
        """
        Prints information about an action that is being performed
        """
        self._print("green", output, newline)

    def error(self, output, newline=False):
        """
        Prints information about an error that has occurred
        """
        self._print("red", output, newline)

    def info(self, output, newline=True):
        """
        Prints information that does not pertain to an action or an error
        """
        self._print("green", output, newline)

    def warning(self, output, newline=True):
        """
        Prints a warning (something that is not an error as of today, but might break in future releases)
        """
        self._print("yellow", output, newline)

    def _print(self, colour, output, newline):
        whitespace = "\n" if newline == True else ""
        print(colored(whitespace + self.prefix + output, color=colour), file=sys.stderr)



from ..infrastructure import DockerUtils, WindowsUtils
from os.path import abspath, dirname, join
import subprocess


class DiagnosticBase(object):
    def getName(self):
        """
        Returns the human-readable name of the diagnostic
        """
        raise NotImplementedError

    def getDescription(self):
        """
        Returns a description of what the diagnostic does
        """
        raise NotImplementedError

    def getPrefix(self):
        """
        Returns the short name of the diagnostic for use in log output
        """
        raise NotImplementedError

    def run(self, logger, args=[]):
        """
        Runs the diagnostic
        """
        raise NotImplementedError

    # Helper functionality for derived classes

    def _printAndRun(self, logger, prefix, command, check=False):
        """
        Prints a command and then executes it
        """
        logger.info(prefix + "Run: {}".format(command), False)
        subprocess.run(command, check=check)

    def _checkPlatformMistmatch(self, logger, containerPlatform):
        """
        Verifies that the user isn't trying to test Windows containers under Windows 10 when in Linux container mode (or vice versa)
        """
        prefix = self.getPrefix()
        dockerInfo = DockerUtils.info()
        dockerPlatform = dockerInfo["OSType"].lower()
        if containerPlatform == "windows" and dockerPlatform == "linux":
            logger.error(
                "[{}] Error: attempting to test Windows containers while Docker Desktop is in Linux container mode.".format(
                    prefix
                ),
                False,
            )
            logger.error(
                "[{}] Use the --linux flag if you want to test Linux containers instead.".format(
                    prefix
                ),
                False,
            )
            raise RuntimeError
        elif containerPlatform == "linux" and dockerPlatform == "windows":
            logger.error(
                "[{}] Error: attempting to test Linux containers while Docker Desktop is in Windows container mode.".format(
                    prefix
                ),
                False,
            )
            logger.error(
                "[{}] Remove the --linux flag if you want to test Windows containers instead.".format(
                    prefix
                ),
                False,
            )
            raise RuntimeError

    def _generateWindowsBuildArgs(
        self, logger, basetagOverride=None, isolationOverride=None
    ):
        """
        Generates the build arguments for testing Windows containers, with optional overrides for base tag and isolation mode
        """

        # Determine the appropriate container image base tag for the host system release unless the user specified a base tag
        buildArgs = []
        hostBaseTag = WindowsUtils.getHostBaseTag()
        baseTag = basetagOverride if basetagOverride is not None else hostBaseTag

        if baseTag is None:
            raise RuntimeError(
                "unable to determine Windows Server Core base image tag from host system. Specify it explicitly using -basetag command-line flag"
            )

        buildArgs = ["--build-arg", "BASETAG={}".format(baseTag)]

        # Use the default isolation mode unless requested otherwise
        dockerInfo = DockerUtils.info()
        isolation = (
            isolationOverride
            if isolationOverride is not None
            else dockerInfo["Isolation"]
        )
        buildArgs += ["--isolation={}".format(isolation)]

        # If the user specified process isolation mode and a different base tag to the host system then warn them
        prefix = self.getPrefix()
        if isolation == "process" and baseTag != hostBaseTag:
            logger.info(
                "[{}] Warning: attempting to use different Windows container/host versions".format(
                    prefix
                ),
                False,
            )
            logger.info(
                "[{}] when running in process isolation mode, this will usually break!".format(
                    prefix
                ),
                False,
            )

        # Set a sensible memory limit when using Hyper-V isolation mode
        if isolation == "hyperv":
            buildArgs += ["-m", "4GiB"]

        return buildArgs

    def _buildDockerfile(self, logger, containerPlatform, tag, buildArgs):
        """
        Attempts to build the diagnostic's Dockerfile for the specified container platform, with the specified parameters
        """

        # Attempt to build the Dockerfile
        prefix = self.getPrefix()
        contextDir = join(
            dirname(dirname(abspath(__file__))),
            "dockerfiles",
            "diagnostics",
            prefix,
            containerPlatform,
        )
        try:
            command = ["docker", "build", "-t", tag, contextDir] + buildArgs
            self._printAndRun(logger, "[{}] ".format(prefix), command, check=True)
            built = True
        except:
            logger.error("[{}] Build failed!".format(prefix))
            built = False

        # Remove any built images, including intermediate images
        logger.action("[{}] Cleaning up...".format(prefix), False)
        if built == True:
            self._printAndRun(logger, "[{}] ".format(prefix), ["docker", "rmi", tag])
        self._printAndRun(
            logger, "[{}] ".format(prefix), ["docker", "system", "prune", "-f"]
        )

        # Report the success or failure of the build
        return built



#!/usr/bin/env python3
import os, sys


def readFile(filename):
    with open(filename, "rb") as f:
        return f.read().decode("utf-8")


def writeFile(filename, data):
    with open(filename, "wb") as f:
        f.write(data.encode("utf-8"))


# Read the build graph XML
buildXml = sys.argv[1]
code = readFile(buildXml)

# Disable HoloLens by default on 4.23.
# On later versions, it is guarded by DefaultWithPlatform and disabled because of HostPlatformOnly=true
code = code.replace(
    'Option Name="WithHoloLens" Restrict="true|false" DefaultValue="$(DefaultWithWindows)"',
    'Option Name="WithHoloLens" Restrict="true|false" DefaultValue="false"',
)

# Enable client and server targets by default in 4.23.0 onwards
code = code.replace(
    'Option Name="WithClient" Restrict="true|false" DefaultValue="false"',
    'Option Name="WithClient" Restrict="true|false" DefaultValue="true"',
)
code = code.replace(
    'Option Name="WithServer" Restrict="true|false" DefaultValue="false"',
    'Option Name="WithServer" Restrict="true|false" DefaultValue="true"',
)

# Write the modified XML back to disk
writeFile(buildXml, code)
